#  Copyright (c) 2024.
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the “Software”), to deal in the
#   Software without restriction,
#   including without limitation the rights to use, copy, modify, merge, publish, distribute,
#   sublicense, and/or sell copies
#   of the Software, and to permit persons to whom the Software is furnished to do so, subject to
#   the following conditions:
#  #
#   The above copyright notice and this permission notice shall be included in all copies or
#   substantial portions of the Software.
#  #
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#   BUT NOT LIMITED TO THE
#   WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
#   EVENT SHALL THE AUTHORS OR
#   COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
#   CONTRACT, TORT OR
#   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#  #
#   This uses QT for some components which has the primary open-source license is the GNU Lesser
#   General Public License v. 3 (“LGPL”).
#   With the LGPL license option, you can use the essential libraries and some add-on libraries
#   of Qt.
#   See https://www.qt.io/licensing/open-source-lgpl-obligations for QT details.

#
#
import os

import yaml
from YMLEditor.data_manager import DataManager


class YamlConfig(DataManager):
    """
    Handles loading and saving YAML files.

    Extends DataManager:
        - Implements YAML specific file _load_data, _save_data

    Inherits DataManager base file handling functionality:
        - load, save, get, set, and undo.

    **Methods**:
    """

    def _load_data(self, f):
        # This will load data from a YAML file
        data = yaml.safe_load(f)

        # Handle case where data is None (empty file) or incorrect format
        if data is None:
            if os.path.exists(self.file_path):
                # File is unreadable as YML
                print(f"Warning: {self.file_path} is not a valid YAML file.")
        return data

    def _save_data(self, f, data):
        # Save data in YAML format
        if data:
            # Save the updated data to the file
            yaml.safe_dump(data, f, sort_keys=False, default_flow_style=False)
        else:
            raise ValueError("_data is None")



#  Copyright (c) 2024. Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated
#  documentation files (the “Software”), to deal in the Software without restriction, including
#  without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
#  Software, and to permit
#  persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or
#  substantial portions of the
#  Software.
#
#  THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE
#  WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
#  EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
#  CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from datetime import date

import pytest

from YMLEditor.yaml_config import YamlConfig

# Define constants for YAML file content
VALID_YAML_FILE = "config.yaml"


@pytest.fixture
def valid_yaml_data():
    return """
---
CRS1: -t_srs epsg:3857
LAYER: B
FILES:
  A: null
  B: WESTMAN.tif
  C: "*_topo.tif"
languages:
  - YAML
  - JAVA
  - XML
name: yaml
rank: 1
born: 2001
published: true
approved: true
date: 1987-05-24
time: 10:00
hash_example: "This string contains # as text"
file_name: -file.txt
"[part1, part2]": This is a complex key with a sequence
quote: ‘YAML is the “best” configuration language’
doublequote: “ Yes, the ‘best’ “
colors:
  - red
  - green
  - blue
person:
  name: Alice
  age: 30
  location: Wonderland
items:
  - id: 1
    name: Item A
  - id: 2
    name: Item B
key_with_empty_value: null
key_with_null_value: null
another_null_value: null
message1: |
  even though it looks like this is a multiline message, it is actually not
message2: |
  this is
  a real multiline
  message
message3: >+
  This block line Will be interpreted as a single line with a newline
  character at the end
vars:
  service1:
    config:
      env: prod
      retries: 3
      version: 4.8
  service2:
    config:
      env: prod
      retries: 3
      version: 4.8
  service3:
    config:
      env: prod
      retries: 3
      version: 4.8
vars2:
  service1:
    config:
      env: prod
      retries: 3
      version: 4.8
  service2:
    config:
      env: prod
      retries: 3
      version: 5
  service3:
    config:
      env: prod
      retries: 3
      version: 4.2
"""


@pytest.fixture
def invalid_yaml_data():
    return """
CRS1: -t_srs epsg:3857
unbalanced brackets: ][
    """


@pytest.fixture
def config():
    return YamlConfig()


@pytest.fixture
def loaded_config(config, valid_yaml_data, tmpdir):
    """Fixture to load configuration from valid YAML data."""
    file_path = tmpdir.join(VALID_YAML_FILE)
    file_path.write(valid_yaml_data)
    config.load(file_path)
    return config


# TEST READ VALUES
def test_read_key(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["CRS1"] == "-t_srs epsg:3857"
    assert loaded_config["LAYER"] == 'B'
    assert loaded_config["FILES"] == {'A': None, 'B': 'WESTMAN.tif', 'C': '*_topo.tif'}


def test_flow_style_sequences(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["colors"] == ["red", "green", "blue"]
    assert loaded_config["languages"] == ["YAML", "JAVA", "XML"]


def test_flow_style_maps(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["person"] == {"name": "Alice", "age": 30, "location": "Wonderland"}


def test_empty_and_null_values(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["key_with_empty_value"] is None
    assert loaded_config["key_with_null_value"] is None
    assert loaded_config["another_null_value"] is None


def test_special_characters(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["quote"] == "‘YAML is the “best” configuration language’"
    assert loaded_config["doublequote"] == "“ Yes, the ‘best’ “"
    assert loaded_config["time"] == 600
    assert loaded_config["date"] == date(1987, 5, 24)
    assert loaded_config["file_name"] == "-file.txt"
    assert loaded_config["hash_example"] == "This string contains # as text"


def test_special_characters2(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["hash_example"] == "This string contains # as text"


def test_multiline_literals1(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["message1"] == ("even though it looks like this is a multiline message, "
                                         "it is actually not\n")


def test_multiline_literals2(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["message2"] == "this is\na real multiline\nmessage\n"


def test_aliases_and_references(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config["vars"]["service1"]["config"]["env"] == "prod"
    assert loaded_config["vars"]["service2"]["config"][
               "env"] == "prod"  # should reference service1 config
    assert loaded_config["vars"]["service3"]["config"][
               "retries"] == 3  # should reference service1 config


def test_read_dot_key(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config.get("FILES.B") == 'WESTMAN.tif'


def test_read_indirect(loaded_config, valid_yaml_data, tmpdir):
    assert loaded_config.get("FILES.@LAYER") == 'WESTMAN.tif'


def test_data_types(loaded_config, valid_yaml_data, tmpdir):
    # Test correct data types and values
    assert isinstance(loaded_config["name"], str)
    assert loaded_config["name"] == "yaml"

    assert isinstance(loaded_config["rank"], int)
    assert loaded_config["rank"] == 1

    assert isinstance(loaded_config["born"], int)
    assert loaded_config["born"] == 2001

    assert isinstance(loaded_config["published"], bool)
    assert loaded_config["published"] is True

    assert isinstance(loaded_config["approved"], bool)
    assert loaded_config["approved"] is True


# TEST SET VALUES
def test_set_and_get_setting(loaded_config, valid_yaml_data, tmpdir):
    loaded_config["NEW_KEY"] = "new_value"
    assert loaded_config["NEW_KEY"] == "new_value"


def test_set_and_get_multi(loaded_config, valid_yaml_data, tmpdir):
    loaded_config["NEW_KEY.A"] = "new_value"
    assert loaded_config["NEW_KEY.A"] == "new_value"


def test_set_and_get_indirect(loaded_config, valid_yaml_data, tmpdir):
    loaded_config.set("FILES.@LAYER", "value2")
    assert loaded_config.get("FILES.@LAYER") == "value2"
    assert loaded_config.get("FILES.B") == "value2"




#  Copyright (c) 2024.
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the “Software”), to deal in the
#   Software without restriction,
#   including without limitation the rights to use, copy, modify, merge, publish, distribute,
#   sublicense, and/or sell copies
#   of the Software, and to permit persons to whom the Software is furnished to do so, subject to
#   the following conditions:
#  #
#   The above copyright notice and this permission notice shall be included in all copies or
#   substantial portions of the Software.
#  #
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#   BUT NOT LIMITED TO THE
#   WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
#   EVENT SHALL THE AUTHORS OR
#   COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
#   CONTRACT, TORT OR
#   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#  #
#   This uses QT for some components which has the primary open-source license is the GNU Lesser
#   General Public License v. 3 (“LGPL”).
#   With the LGPL license option, you can use the essential libraries and some add-on libraries
#   of Qt.
#   See https://www.qt.io/licensing/open-source-lgpl-obligations for QT details.

from functools import partial
from typing import Union, List

# Use PySide for imports and fall back to PyQt
try:
    from PySide6.QtCore import Qt
    from PySide6.QtWidgets import QWidget, QLabel, QComboBox, QLineEdit, QSizePolicy, QTextEdit, QCheckBox, QSlider, \
        QDoubleSpinBox
except ImportError:
    from PyQt6.QtCore import Qt
    from PyQt6.QtWidgets import QWidget, QLabel, QComboBox, QLineEdit, QSizePolicy, QTextEdit, QCheckBox, QSlider, \
        QDoubleSpinBox


from YMLEditor.structured_text import to_text, data_type, parse_text, _ast_dictionary


class ItemWidget(QWidget):
    """
    A configurable widget for displaying and editing a single field from a config file.

    - Supports various widget types including editable text, combo boxes, etc.  See readme for details.
    - All user edits are validated and synchronized with the config data.

    **Methods**:
    """

    def __init__(
            self, config, widget_type, initial_value, combo_rgx, callback, width=50, key=None,
            text_edit_height=60, verbose=3, error_style="color: Orange;", style=None
    ):
        """
        Init

        Args:
            config(Config): Configuration handler to synchronize data.
            widget_type (str): Type of widget to create
                ("text_edit", "line_edit", "read_only", "combo", "label").
            initial_value (str): Initial value to populate the widget.
            combo_rgx (Union[List[str], str]): Dropdown options for combo boxes or
                regex for validating text fields or tuple with min/max for a slider
            callback (callable): Function to call when the widget value changes.
            width (int, optional): Fixed width for the widget. Defaults to 50.
            key (str, optional): Key for linking the widget to the config data.
            text_edit_height (int, optional): Height for text edit widgets. Defaults to 90.
            verbose (int, optional): Verbosity level. 0=silent, 3=warnings, 4=information.
                Defaults to 3.
            error_style (str, optional): style for indicating an error.
            style (str) : style for the widget
        """
        super().__init__()

        self._error_style = error_style
        self._rgx = None
        self._widget_type = widget_type
        self._callback = callback
        self._key = key
        self._config = config
        self._is_valid = False
        self._data_type = None
        self._verbose = verbose

        self._create_widget(widget_type, initial_value, combo_rgx, width, text_edit_height, style)

    def display(self):
        """
        Load and display our field from the config data.
        Prints a warning if our key is not found in config data.
        """
        key, val = None, None
        try:
            if self.widget:
                key = self.widget.objectName()
                if key:
                    val = self._config.get(key) or ""
                    if not self._data_type:
                        self._data_type = data_type(val)

                    self.set_text(self.widget, val)

        except Exception as e:
            key = key or "None"
            val = val or "None"
            self.warn(f"WARN:  key '{key}': {e} val '{val}'")

    def _on_widget_changed(self, widget, *args):
        """
        Handle changes to the widget's value: validate text. If valid,
        update the config data. Set style appropriately.

        Args:
            widget (QWidget): The widget whose value was changed.
        """
        key = widget.objectName()
        text = get_text(widget)

        # Ensure text is valid and properly formatted
        if isinstance(text, str) and self._data_type:
            text = text.strip()  # Remove surrounding whitespace
            try:
                if self._data_type == dict:
                    text = _ast_dictionary(text)
                elif self._data_type == list:
                    # Add enclosing brackets if not present and content is non-empty
                    if text and (not text.startswith("[") or not text.endswith("]")):
                        text = f"[{text}]"
            except Exception as e:
                self.warn(f"Widget get text error. Text is '{text}'")
                self.set_error_style(widget)
                return

        # Validate the text and parse it
        invalid, data_value = parse_text(text, self._data_type, self._rgx)

        # Update config and apply styles based on validation
        if invalid:
            self.warn(f"parse error for {text}")
            self.set_error_style(widget)
        else:
            try:
                self._config.set(key, data_value)
                self.set_normal_style(widget)
            except Exception as e:
                self.set_error_style(widget)
                self.info(f"Error setting {text}")
            self._callback(key, text)

    def set_error_style(self, widget, message=None):
        """
        Apply an error style to the widget.

        Args:
            widget (QWidget): The widget to style.
            message (str, optional): Optional error message to display.
        """
        if not widget.property("originalStyle"):
            name = widget.objectName()
            widget.setProperty("originalStyle", widget.styleSheet())

        widget.setStyleSheet(self._error_style)
        if message:
            widget.setText(message)

    def set_normal_style(self, widget):
        """
        Restore the widget's default style.

        Args:
            widget (QWidget): The widget to restore.
        """
        original_style = widget.property("originalStyle")
        widget.setStyleSheet(original_style)

    def _create_widget(self, widget_type, initial_value, options, width, text_edit_height, style):
        """
        Create a specific type of widget based on the provided parameters (private).

        Args:
            widget_type (str): The type of widget to create.
            initial_value (str): The initial value for the widget.
            options (Union[List, str], optional): Regex for text fields, widget options for others.
            width (int): Width of the widget.
            text_edit_height (int): Height for text edit widgets.
            style (str): Style for the widget.
        """
        # Validate options based on widget type
        if widget_type in ["combo", "slider", "spinbox"] and not isinstance(options, list):
            raise ValueError(f"Options for '{widget_type}' must be a list. Got: {type(options).__name__}")

        if widget_type == "slider":
            if len(options) != 2 or not all(isinstance(i, int) for i in options):
                raise ValueError(
                    f"Options for 'slider' must be a list of two integers [min, max]. Got: {options}"
                )

        if widget_type == "spinbox":
            if len(options) != 4 or not (
                isinstance(options[0], (int, float)) and
                isinstance(options[1], (int, float)) and
                isinstance(options[2], (int, float)) and
                isinstance(options[3], int)
            ):
                raise ValueError(
                    f"Options for 'spinbox' must be a list of [min (float), max (float), step (float), precision (int)]. "
                    f"Got: {options}"
                )

        if widget_type == "combo" and not all(isinstance(i, str) for i in options):
            raise ValueError(
                f"Options for 'combo' must be a list of strings. Got: {options}"
            )

        # Create Widget
        if widget_type == "combo":
            self.widget = QComboBox()
            self.widget.addItems(options)
            self.set_text(self.widget, initial_value)
            self.widget.currentIndexChanged.connect(partial(self._on_widget_changed, self.widget))
        elif widget_type == "text_edit":
            self.widget = QTextEdit(str(initial_value))
            self.widget.setFixedHeight(text_edit_height)
            self.widget.setAcceptDrops(False)
            self._rgx = options
            self.widget.textChanged.connect(partial(self._on_widget_changed, self.widget))
        elif widget_type == "line_edit":
            self.widget = QLineEdit(str(initial_value))
            self.widget.setAcceptDrops(False)
            self._rgx = options
            self.widget.textChanged.connect(partial(self._on_widget_changed, self.widget))
        elif widget_type == "read_only":
            self.widget = QLineEdit(str(initial_value))
            self.widget.setReadOnly(True)
        elif widget_type == "label":
            self.widget = QLabel()
        elif widget_type == "checkbox":
            self.widget = QCheckBox()
            self.set_text(self.widget, initial_value)
            self.widget.stateChanged.connect(partial(self._on_widget_changed, self.widget))
        elif widget_type == "slider":
            self.widget = QSlider(Qt.Orientation.Horizontal)
            self.widget.setMinimum(options[0])
            self.widget.setMaximum(options[1])
            self.set_text(self.widget, initial_value)
            self.widget.valueChanged.connect(partial(self._on_widget_changed, self.widget))
        elif widget_type == "spinbox":
            self.widget = QDoubleSpinBox()
            self.widget.setMinimum(options[0])
            self.widget.setMaximum(options[1])
            self.widget.setSingleStep(options[2])
            self.widget.setDecimals(options[3])
            self.set_text(self.widget, initial_value)
            self.widget.valueChanged.connect(partial(self._on_widget_changed, self.widget))
        else:
            raise TypeError(f"Unsupported widget type: {widget_type} for {self._key}")

        if widget_type not in ["label"]:
            self.widget.setObjectName(self._key)

        if style:
            self.widget.setStyleSheet(style)

        self.widget.setProperty("originalStyle", self.widget.styleSheet())
        if isinstance(self.widget, QLineEdit):
            self.widget.setFixedWidth(width)
        else:
            self.widget.setMinimumWidth(width)

        self.widget.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)


    def set_text(self, widget, data):
        """
        Update the widget's value with the provided value.

        Args:
            widget (QWidget): The widget to update.
            data (str, int, float, bool, or dict): The data to update in the widget.
        """
        self.widget.blockSignals(True)

        if isinstance(widget, QComboBox):
            widget.setCurrentText(str(data))
        elif isinstance(widget, QLineEdit):
            if self._widget_type == "read_only":
                # Remove quotes, braces, and square brackets from str(data)
                cleaned_data = str(data).strip("[]{}").replace("'", "").replace('"', "")
                widget.setText(cleaned_data)
            else:
                widget.setText(str(data))
        elif isinstance(widget, QTextEdit):
                widget.setPlainText(str(data))
        elif isinstance(widget, QCheckBox):
            widget.setChecked(data == '1' or data == 'True')
        elif isinstance(widget, QSlider):
            if data:
                widget.setValue(int(data))
        elif isinstance(widget, QDoubleSpinBox):
            if data:
                widget.setValue(float(data))
        else:
            self.widget.blockSignals(False)
            raise TypeError(f"Unsupported widget type for setting value: {type(widget)}")

        self.widget.blockSignals(False)

    def warn(self, text):
        if self._verbose > 2:
            print(f"Warning: {text}")

    def info(self, text):
        if self._verbose > 3:
            print(f"Info: {text}")

def get_text(widget):
    """
    Retrieve the value from a widget.

    Args:
        widget (QWidget): The widget to retrieve the value from.

    Returns:
        str: The current value of the widget as a string.
    """
    if isinstance(widget, QComboBox):
        return widget.currentText()
    elif isinstance(widget, QTextEdit):
        return widget.toPlainText()
    elif isinstance(widget, QCheckBox):
        return "1" if widget.isChecked() else "0"
    elif isinstance(widget, QSlider):
        return str(widget.value())
    elif isinstance(widget, QDoubleSpinBox):
        # Round to the precision set in the QDoubleSpinBox
        precision = widget.decimals()
        value = round(widget.value(), precision)
        return f"{value:.{precision}f}"  # Format to ensure consistent decimal places
    elif hasattr(widget, "text"):
        return widget.text()
    else:
        raise TypeError(f"Unsupported widget type: {type(widget)}")



# yaml_syntax_highlighter.py
from PySide6.QtCore import QRegularExpression
from PySide6.QtGui import QColor, QTextCharFormat, QFont, QSyntaxHighlighter

class YamlHighlighter(QSyntaxHighlighter):
    """
    A syntax highlighter for YAML files that can also highlight a single line for errors.
    """
    def __init__(self, parent=None):
        super().__init__(parent)

        self.error_line = -1  # Line number (1-based) of the error

        # --- Standard YAML Syntax Rules ---
        self.highlighting_rules = []

        # Rule for keys (e.g., "key:")
        key_format = QTextCharFormat()
        key_format.setForeground(QColor("#9CDCFE")) # Light Blue
        key_format.setFontWeight(QFont.Bold)
        self.highlighting_rules.append((QRegularExpression(r"^\s*([a-zA-Z0-9_-]+):"), key_format))

        # Rule for comments (e.g., "# a comment")
        comment_format = QTextCharFormat()
        comment_format.setForeground(QColor("#6A9955")) # Green
        self.highlighting_rules.append((QRegularExpression(r"#[^\n]*"), comment_format))

        # Rule for string values in quotes
        string_format = QTextCharFormat()
        string_format.setForeground(QColor("#CE9178")) # Orange
        self.highlighting_rules.append((QRegularExpression(r"(\".*\"|\'.*\')"), string_format))

        # --- Error Highlighting Rule ---
        self.error_format = QTextCharFormat()
        self.error_format.setBackground(QColor("#5A1D1D")) # Dark Red background

    def set_error_line(self, line_number: int | None):
        """
        Sets the line number to be highlighted as an error.
        Pass None or -1 to clear the error highlight.
        """
        self.error_line = line_number if line_number is not None else -1
        # We must re-highlight the entire document for the change to take effect.
        self.rehighlight()

    def highlightBlock(self, text: str):
        """
        This method is called by Qt for each line of text to apply formatting.
        """
        # --- Check for Error First ---
        # blockNumber() is 0-based, our line numbers are 1-based.
        current_line_number = self.currentBlock().blockNumber() + 1
        if self.error_line == current_line_number:
            self.setFormat(0, len(text), self.error_format)
            return # Stop further processing for this line if it's an error

        # --- Apply Standard Syntax Highlighting ---
        for pattern, format in self.highlighting_rules:
            expression = QRegularExpression(pattern)
            it = expression.globalMatch(text)
            while it.hasNext():
                match = it.next()
                self.setFormat(match.capturedStart(), match.capturedLength(), format)


from unittest.mock import patch

import pytest

from YMLEditor.data_manager import DataManager


class TestDataManager:
    """Test suite for the DataManager abstract base class."""

    @pytest.fixture
    def data_manager(self):
        """Fixture to provide a subclass of DataManager for testing."""

        class TestDataManager(DataManager):
            def _load_data(self, f):
                # Mock load implementation
                return {"key": "value"}

            def _save_data(self, f, data):
                # Mock save implementation
                pass

        return TestDataManager(verbose=1)

    def test_set_get(self, data_manager):
        # 1. Initialize "key" to "value" and verify
        data_manager._data = {"key": "value"}
        assert data_manager.get("key") == "value", "Failed to set and retrieve 'key' as 'value'"

        # 2. Set "key2" to None and verify
        data_manager.set("key2", None)
        assert data_manager.get("key2") is None, "Failed to set and retrieve 'key2' as None"

        # 3. Set "key" to "value" again and verify
        data_manager.set("key2", "value")
        assert data_manager.get("key2") == "value", \
            "Failed to set and retrieve 'key2' as 'value' again"

    def test_initialization(self, data_manager):
        """Test the initialization of DataManager."""
        assert data_manager.verbose == 1
        assert data_manager._data is None
        assert data_manager.file_path is None
        assert data_manager.unsaved_changes is False
        assert data_manager.snapshots == []

    def test_save_success(self, data_manager, tmp_path):
        """Test saving data successfully."""
        file_path = tmp_path / "test_file.json"
        data_manager.file_path = str(file_path)
        data_manager._data = {"key": "value"}
        data_manager.unsaved_changes = True

        with patch.object(data_manager, '_save_data', return_value=None):
            assert data_manager.save()
            assert not data_manager.unsaved_changes

    def test_save_without_file_path(self, data_manager):
        """Test saving without a file path."""
        data_manager._data = {"key": "value"}
        with pytest.raises(ValueError):
            data_manager.save()

    def test_save_without_data(self, data_manager):
        """Test saving without data."""
        data_manager.file_path = "test_file.json"
        with pytest.raises(ValueError):
            data_manager.save()

    def test_snapshot_push(self, data_manager):
        """Test pushing snapshots."""
        data_manager.init_data({"key": "value1"})
        data_manager.snapshot_push()
        assert len(data_manager.snapshots) == 2

        # Test max snapshot limit
        for i in range(data_manager.max_snapshots):
            data_manager.snapshot_push()
        assert len(data_manager.snapshots) == data_manager.max_snapshots

    def test_snapshot_undo(self, data_manager):
        """Test undo functionality."""
        data_manager.init_data({"key": "value1"})
        data_manager.snapshot_push()
        data_manager.set("key", "value2")
        data_manager.snapshot_push()

        data_manager.snapshot_undo()
        assert data_manager._data == {"key": "value2"}

        data_manager.snapshot_undo()
        assert data_manager._data == {"key": "value1"}

    def test_get_and_set(self, data_manager):
        """Test get and set operations."""
        data_manager.init_data({"key": "value"})
        assert data_manager.get("key") == "value"

        data_manager.set("key", "new_value")
        assert data_manager.get("key") == "new_value"

    def test_add_proxy(self, data_manager):
        """Test adding a proxy file."""
        proxy_file = "proxy_file"
        keys = ["key1", "key2"]

        data_manager.register_proxy_file(proxy_file, keys)
        assert data_manager._get_proxy("key1") == proxy_file
        assert data_manager._get_proxy("key2") == proxy_file

        # Adding duplicate key should raise ValueError
        with pytest.raises(ValueError):
            data_manager.register_proxy_file(proxy_file, ["key1"])

    def test_delete(self, data_manager):
        """Test delete operation."""
        data_manager.init_data({"key1": "value1", "key2": "value2"})
        data_manager.delete("key1")
        assert "key1" not in data_manager._data

    def test_insert(self, data_manager):
        """Test insert operation."""
        data_manager.init_data([])
        data_manager.insert(0, "item1")
        assert data_manager.get(0) == "item1"



from pathlib import Path
from typing import List, Dict, Any, Type, Optional, Tuple

from cerberus import Validator
import yaml


class _Loader(yaml.SafeLoader):
    """A custom `yaml.SafeLoader` that adds support for the `!include` tag."""
    def __init__(self, stream):
        """Initializes the loader and sets the root for relative includes."""
        try:
            self._root = Path(stream.name).parent
        except AttributeError:
            self._root = Path.cwd()
        super().__init__(stream)

def _include_constructor(loader: _Loader, node: yaml.Node) -> Dict:
    """Loads and parses a YAML file specified by an !include tag."""
    include_path = loader._root / Path(loader.construct_scalar(node))
    if not include_path.is_file():
        raise FileNotFoundError(f"Included file '{include_path}' not found.")
    with include_path.open('r', encoding='utf-8') as f:
        return yaml.load(f, Loader=_Loader) or {}

yaml.add_constructor('!include', _include_constructor, Loader=_Loader)


class ConfigLoader:
    """
    A reusable class to load, validate, and provide rich error reporting for YAML files.
    """
    def __init__(
            self, schema: Dict = None,
            error_modifications: Optional[List[Dict]] = None,
            validator_class: Type[Validator] = Validator
    ):
        """Initializes the ConfigLoader."""
        self.schema = schema
        self.validator = validator_class(self.schema) if self.schema else None

        # Rules for text modifications in Cerberus error messages
        self.error_modifications = error_modifications or ERROR_MODIFICATION_RULES

    def read(
            self,
            config_file: Path,
            allow_unknown: bool = False,
            root_node: str | None = None,
            block_descriptor: str | None = None,
            normalize: bool = False
    ) -> dict:
        """
          docstring unchanged...
        """
        if not isinstance(config_file, Path):
            config_file = Path(config_file)
        if not config_file.exists():
            raise FileNotFoundError(f"Configuration file not found: {Path.cwd()}/{config_file}")

        try:
            with config_file.open("r", encoding="utf-8") as f:
                yml_config = yaml.load(f, Loader=_Loader) or {}

        except yaml.YAMLError as e:
            # ---  uses the unified translation engine ---
            raw_error = str(e)
            problem, suggestion = self._translate_error_message(raw_error)

            error_message = (
                f"   Error: YAML syntax error in '{config_file}':\n"
                f"    {problem}\n"
                f"    {suggestion}\n\n"
                f"    Details:\n{raw_error}"
            )
            raise ValueError(error_message) from e

        if self.validator:
            # --- Set allow_unknown directly for this read operation ---
            self.validator.allow_unknown = allow_unknown

            if not self.validator.validate(yml_config):
                formatted_errors = "\n".join(self._format_yml_errors(
                    self.validator.errors,
                    yml_config,
                    root_node=root_node,
                    block_descriptor=block_descriptor
                ))
                raise ValueError(f"  ❌ Error: Configuration file '{config_file}' has errors:\n{formatted_errors}")

            return self.validator.normalized(yml_config) if normalize else yml_config

        # If no validator, just return the raw config
        return yml_config

    def _format_yml_errors(
            self,
            errors: Dict,
            data: Dict,
            prefix: str = "",
            root_node: str | None = None,
            block_descriptor: str | None = None
    ) -> List[str]:
        messages = []
        for field, issues in errors.items():
            full_path_str = f"{prefix}.{field}" if prefix else str(field)
            full_path_list = full_path_str.split('.')

            if isinstance(issues, list):
                for issue in issues:
                    if isinstance(issue, dict):
                        messages.extend(self._format_yml_errors(
                            issue, data, prefix=full_path_str,
                            root_node=root_node, block_descriptor=block_descriptor
                        ))
                    else:
                        location_messages = []
                        parent_data = None

                        if root_node and block_descriptor and full_path_list[0] == root_node and len(full_path_list) > 1:
                            block_path = full_path_list[:2]
                            block_data = self._get_value_from_path(data, block_path)
                            if isinstance(block_data, dict) and block_descriptor in block_data:
                                descriptor_value = block_data[block_descriptor]
                                location_messages.append(f"In the block with {block_descriptor} '{descriptor_value}'.")

                        if len(full_path_list) > 0:
                            parent_path = full_path_list[:-1]
                            parent_data = self._get_value_from_path(data, parent_path)
                            if isinstance(parent_data, list):
                                error_index = int(full_path_list[-1])
                                last_key = full_path_list[-2] if len(full_path_list) > 1 else 'list'
                                location_messages.append(f"The error is in the {self._format_human_readable_path(full_path_list[-1:])} in the '{last_key}' list.")
                                if error_index > 0:
                                    preceding = parent_data[max(0, error_index - 1):error_index]
                                    if preceding:
                                        location_messages.append(f"It comes right after '{preceding[0]}'.")

                        if not location_messages:
                            location_messages.append(f"The error is in {self._format_human_readable_path(full_path_list)}.")

                        problem, suggestion = self._translate_cerberus_issue(issue, full_path_list, parent_data)
                        found_value = self._get_value_from_path(data, full_path_list)

                        formatted_message = (
                            f"\n    Problem: {problem}"
                            f"\n    Found Value: {repr(found_value)}"
                            f"\n    {' '.join(location_messages)}"
                            f"\n    {suggestion}"
                            f"\n\n    "
                            f"\n      Item: {' / '.join(full_path_list)}"
                            f"\n      {issue}"
                        )
                        messages.append(formatted_message)
        return messages

    def _translate_error_message(self, error_string: str) -> tuple[str, str]:
        """
        Translates a raw error string using the configurable translation rules.
        """
        for rule in self.error_modifications:
            trigger_phrase = rule.get("trigger_phrase", "")
            if trigger_phrase and trigger_phrase in error_string:
                # Found a matching rule.
                return rule["problem"], rule["suggestion"]

        # Generic fallback if no specific rule matched.
        problem = f"An unknown error occurred: {error_string}"
        suggestion = "Please check the file syntax and the documentation."
        return problem, suggestion

    def _translate_cerberus_issue(self, issue: str, path: List[str], parent_data: Any) -> tuple[str, str]:
        """
        Translates a single Cerberus error, modifying text phrases and enhancing it with contextual info.
        """
        field_name = path[-1] if path else "a setting"

        # Call the unified translator first to get the base message.
        problem, suggestion = self._translate_error_message(issue)

        # Format the base message with the specific field name.
        problem = problem.format(field_name=field_name)
        suggestion = suggestion.format(field_name=field_name)

        # Enhance "unallowed value" errors with specific details from the schema.
        if "unallowed value" in issue:
            schema_rule = self._get_schema_rule_from_path(self.validator.schema, path)
            allowed_values_str = ""
            if schema_rule and 'allowed' in schema_rule:
                allowed_values_str = ", ".join(f"'{v}'" for v in schema_rule['allowed'])

            if allowed_values_str:
                found_value = issue.split("'")[1] if "'" in issue else "an invalid value"
                problem = f"{problem} Expected: {allowed_values_str}. Found: '{found_value}'."

        return problem, suggestion

    def _get_schema_rule_from_path(self, schema: Any, path: List[str]) -> Optional[Dict]:
        """Navigates the schema dictionary using a path list to find a rule."""
        current_schema_item = schema

        # Handle cases where the schema is wrapped in a list. This makes the
        # function robust against this specific formatting issue.
        if isinstance(current_schema_item, list) and len(current_schema_item) > 0:
            current_schema_item = current_schema_item[0]

        for key in path:
            if not hasattr(current_schema_item, 'get'):
                return None  # Cannot traverse further if it's not a dictionary like

            current_schema_item = current_schema_item.get(key)
            if current_schema_item is None:
                return None

            # For list validation, the rule is nested under 'schema'
            if isinstance(current_schema_item, dict) and 'schema' in current_schema_item:
                current_schema_item = current_schema_item['schema']

        return current_schema_item if isinstance(current_schema_item, dict) else None

    @staticmethod
    def _get_value_from_path(data: Any, path: List[str]) -> Any:
        current = data
        for key in path:
            try:
                if isinstance(current, list) and key.isdigit():
                    current = current[int(key)]
                elif isinstance(current, dict):
                    current = current[key]
                else:
                    return None
            except (KeyError, IndexError, TypeError):
                return None
        return current

    @staticmethod
    def _parse_type_error(issue: str) -> tuple[str, str] | None:
        trigger_words = {'must', 'be', 'of', 'type'}
        type_map = {
            'string': ('text', "Ensure the value is text. If using a number or a word like 'yes', surround it with quotes (e.g., 'yes')."),
            'integer': ('an integer', "Use a number without decimals (e.g., 20 or -5)."),
            'float': ('a number', "Use a number. Decimals are allowed (e.g., 20.7)."),
            'number': ('a number', "Use a number. Decimals are allowed (e.g., 20.7)."),
            'boolean': ('a true/false value', "Use one of these values: true, false, yes, no, on, off."),
            'list': ('a list of items', 'Format the value as a list, with each item on a new line preceded by a hyphen (e.g., "- item1").'),
            'dict': ('a collection of key-value settings', "Format the value as a set of 'key: value' pairs.")
        }
        words = set(issue.lower().split())
        if not trigger_words.issubset(words): return None
        for technical_type, (human_name, suggestion) in type_map.items():
            if technical_type in words: return human_name, suggestion
        return None

    def _format_human_readable_path(self, path_list: List[str]) -> str:
        parts = []
        for key in path_list:
            if key.isdigit():
                human_index = int(key) + 1
                parts.append(f"the {self._get_ordinal(human_index)} item")
            else:
                parts.append(f"the '{key}' section")
        if len(parts) > 1: return f"{' in '.join(parts[::-1])}"
        return parts[0] if parts else "the top level"

    @staticmethod
    def _get_ordinal(n):
        if 11 <= (n % 100) <= 13: suffix = 'th'
        else: suffix = ['th', 'st', 'nd', 'rd', 'th'][min(n % 10, 4)]
        return str(n) + suffix


ERROR_MODIFICATION_RULES = [
    # --- Rules for modifying  PYAML error text ---

    {
        "key": "YAML_INDENTATION_ERROR",
        "trigger_phrase": "expected <block end>, but found '?'",
        "problem": "This is a YAML indentation error.",
        "suggestion": (
            "Keys within a list item (like 'description:' or 'contains:') "
            "must be indented to the same level, directly under the hyphen ('-') "
            "that starts the list item."
        )
    },
    {
        "key": "YAML_QUOTED_SCALAR",
        "trigger_phrase": "while scanning a quoted scalar",
        "problem": "Invalid syntax in quoted text.",
        "suggestion": "Please check for syntax errors inside a string, such as unescaped characters or missing end-quote."
    },
    {
        "key": "YAML_QUOTED_SCALAR",
        "trigger_phrase": "while scanning a quoted scalar",
        "problem": "Invalid syntax in quoted text.",
        "suggestion": "Please check for syntax errors inside a string, such as unescaped characters or missing end-quote."
    },
    {
        "key": "YAML_UNEXPECTED_END",
        "trigger_phrase": "unexpected end of stream",
        "problem": "The file ended unexpectedly while parsing a value.",
        "suggestion": "This is often caused by a missing closing quote (' or \") or a misplaced colon (:)."
    },
    # --- Rules for modifying Cerberus  error text ---
    {
        "key": "CONFIG_TYPE_MISMATCH",
        "field_name": "config_type",
        "trigger_phrase": "unallowed value",
        "problem": "The config_type for this config file is incorrect",
        "suggestion": "Please check your build configuration to ensure the correct file is being passed for this parameter.",
    },
    {
        "key": "REQUIRED_FIELD",
        "trigger_phrase": "required field",
        "problem": "The required setting '{field_name}' is missing.",
        "suggestion": "Please add '{field_name}: <value>' to this section.",
    },
    {
        "key": "NULL_VALUE",
        "trigger_phrase": "null value",
        "problem": "The setting '{field_name}' is empty.",
        "suggestion": "Please add a value for '{field_name}'.",
    },
    {
        "key": "UNKNOWN_FIELD",
        "trigger_phrase": "unknown field",
        "problem": "The setting '{field_name}' is not allowed here.",
        "suggestion": "Please remove it or check for a typo.",
    },
]


# Configuration file for the Sphinx documentation builder.
#
# To build .rst files for YMLEditor modules:
# Go to the projectroot folder (YMLEditor)
# sphinx-apidoc -o docs/source YMLEditor

import os
import sys

# Include the project root and module paths
sys.path.insert(0, os.path.abspath('../..'))

# Project Information
project = 'YMLEditor'

# Sphinx Extensions
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'myst_parser',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
autoclass_content = 'both'

# HTML Theme
html_theme = "furo"
html_static_path = ['_static']



"""
This provides routines for formatting data structures as text for user editing and then parsing
the edited text back into a data structure.

- Converts data structures to text for display or editing (`to_text`).
- Parses text back into structured data (`parse_text`).
- Supports simple dict, list, and scalar (int, float, bool, etc.)

"""
#   Copyright (c) 2024.
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the “Software”), to deal in the
#   Software without restriction,
#   including without limitation the rights to use, copy, modify, merge, publish, distribute,
#   sublicense, and/or sell copies
#   of the Software, and to permit persons to whom the Software is furnished to do so, subject to
#   the following conditions:
#
#   The above copyright notice and this permission notice shall be included in all copies or
#   substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#   BUT NOT LIMITED TO THE
#   WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
#   EVENT SHALL THE AUTHORS OR
#   COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
#   CONTRACT, TORT OR
#   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#
#   This uses QT for some components which has the primary open-source license is the GNU Lesser
#   General Public License v. 3 (“LGPL”).
#   With the LGPL license option, you can use the essential libraries and some add-on libraries
#   of Qt.
#   See https://www.qt.io/licensing/open-source-lgpl-obligations for QT details.

import ast
from datetime import date, datetime
import re


def to_text(item, normalize=False):
    """
    Converts a Python object into a string representation that is compatible with
    `ast.literal_eval`.
    This function handles common Python types including nested hierarchies and ensures proper
    formatting for compatibility with evaluation and representation needs.

    Args:
        item: The data structure to format. Supported types include:
            - `dict`: Converted into a string with properly formatted key-value pairs.
            - `list`: Converted into a string representation of a list.
            - `str`: Encased in single quotes, ensuring compatibility with evaluators.
            - `int` or `float`: Directly converted into their string representations.
            - `bool`: Represented as `True` or `False` in the string.
            - `None`: Represented as the string ``.
            - `date`: Converted to an ISO-formatted string enclosed in single quotes.

        normalize (bool): If `True`, applies normalization rules to standardize strings for
        certain cases.  Ensures proper escaping of characters in strings using `repr` when
        normalizing.

    Returns:
        str: A string representation of the object, formatted to be compatible with
        `ast.literal_eval`.

    Raises:
        TypeError: If the input object is of an unsupported type.

    Notes:
        - Nested structures (e.g., lists within dictionaries) are handled recursively.
        - The `normalize` flag affects how strings are represented, enabling additional
        standardization when needed.
    """
    # Handle dictionaries by converting each key-value pair into a formatted string
    if isinstance(item, dict):
        formatted_pairs = [f"'{key}': {to_text(val if val is not None and val != '' else '', True)}"
            for key, val in item.items()]
        result = "{" + ", ".join(formatted_pairs) + "}"
        return result

    # Handle lists by converting each element into its formatted string representation
    elif isinstance(item, list):
        formatted_elements = [to_text(element, True) for element in item]
        return "[" + ", ".join(formatted_elements) + "]"

    # Handle booleans by mapping them to their Python string equivalents
    elif isinstance(item, bool):
        return "True" if item else "False"

    # Handle numbers (int and float) by converting directly to their string representation
    elif isinstance(item, (int, float)):
        return str(item)

    # Handle strings, optionally normalizing if the flag is set
    elif isinstance(item, str):
        if normalize:
            # Use `repr` to ensure proper quoting and escaping of special characters, e.g., '\n',
            # '\t'
            return repr(item)
        # Default behavior, return as-is
        return f"{item}"

    # Handle NoneType as the string ""
    elif item is None:
        return "''"

    # Handle date objects by formatting them into ISO 8601 strings wrapped in single quotes
    elif isinstance(item, date):
        return f"'{item.isoformat()}'"

    # Handle tuples
    elif isinstance(item, tuple):
        formatted_elements = [to_text(element, True) for element in item]
        return "(" + ", ".join(formatted_elements) + ("," if len(item) == 1 else "") + ")"
    # Raise an error for unsupported types
    else:
        raise TypeError(f"Unsupported type: {type(item).__name__}")


DEFAULT_FALLBACKS = {
    int: -9999,  # Default fallback for integers
    float: -9999.9,  # Default fallback for floats
    bool: False,  # Default fallback for booleans
    str: "ERROR",  # Default fallback for strings
    date: date(1900, 1, 1),  # Default fallback for dates
    dict: None,  # Default fallback for dictionaries
    list: None,  # Default fallback for lists
    tuple: None,
}

def get_default_fallbacks():
    """
    Retrieve the DEFAULT_FALLBACKS dictionary.
    DEFAULT_FALLBACKS is a dictionary that provides default fallback values for
    various data types (e.g., int, float, str) when text cannot be parsed correctly.

    Defaults:
        - `int`: -9999
        - `float`: -9999.9
        - `bool`: False
        - `str`: "ERROR"
        - `date`: datetime.date(1900, 1, 1)
        - `dict`, `list`, `tuple`: None

    Returns:
        dict: Dictionary of fallback values for parsing failures.
    """
    return DEFAULT_FALLBACKS


def parse_text(text, target_type, rgx=None, fallbacks=None):
    """
    Parse a string into a Python object (e.g., dictionary, list, scalar, or nested structure).
    The input string must be in a format compatible with Python's `ast.literal_eval`.

    Args:
        text (str): The input string to parse.
        target_type (type): The desired type of the output (e.g., int, float, str, bool, date,
        etc.).
        rgx: Optional parameter for additional handling.
        fallbacks: Override default fallbacks.

    Returns:
        tuple: (error_flag, result)
            - error_flag (bool): Indicates if parsing errors or type mismatches occurred.
            - result: The parsed object if successful, or a fallback value if parsing fails.
    Raises:
        None: Any exceptions during parsing are handled internally and result in a fallback value.

    Fallback Handling:
        If parsing fails or the result does not match the desired target type:
        - The function assigns a default fallback value from `DEFAULT_FALLBACKS` based on the
          specified `target_type`.
        - If `target_type` is not in `DEFAULT_FALLBACKS`, `None` is returned as the fallback value.
        - The DEFAULT_FALLBACKS can be overridden with `fallbacks`.

    Notes:
        - Parsing is performed using `ast.literal_eval` for security and compatibility.
        - Nested structures (e.g., lists or dictionaries) are handled recursively.
        - Scalar values are converted to the `target_type` if specified.
    """
    # Use the provided fallbacks or default to `DEFAULT_FALLBACKS` if not provided
    fallbacks = fallbacks or DEFAULT_FALLBACKS

    # Validate against rgx
    if rgx:
        info(f"check regex: val={text} rgx={rgx}")
        valid = validate_text(text, rgx)
        if not valid:
            info(f"regex validation failed for '{text}'")
            if target_type is str:
                # Strings are always returned as-is
                return True, text

            # Assign fallback value based on the target type
            value = fallbacks.get(target_type, None)
            return True, value

    info(f"parse text: {text} target_type={target_type}")

    if target_type is str:
        # Strings are always returned as-is
        return False, text

    # Attempt parsing and initial type matching
    error_flag, value = _parse_text(text, target_type, rgx)

    # If parsing fails or the result does not match the target type
    if error_flag or (target_type and not isinstance(value, target_type)):
        # Assign fallback value based on the target type
        value = fallbacks.get(target_type, None)
        error_flag = True

    return error_flag, value


def _parse_text(text, target_type=None, rgx=None):
    """
    Internal recursive function to parse a string into a Python object (e.g., dictionary, list,
    or scalar).

    Args:
        text (str): The input string to parse.
        target_type (type, optional): Desired data type for the result.
        rgx: Optional parameter for additional handling (not implemented in this version).

    Returns:
        tuple: (error_flag, result)
            - error_flag (bool): True if parsing errors or mismatches occurred, False otherwise.
            - result: The parsed object, or None if parsing fails.

    Notes:
        - Handles dictionaries and lists recursively.
        - Uses `_convert_value` to ensure scalar values match the desired target type.
    """
    try:
        # Attempt parsing the string with `ast.literal_eval`
        info(f"parsing <{text}>")
        result = ast.literal_eval(text)

        # Handle dictionaries by recursively parsing keys and values
        if isinstance(result, dict):
            parsed_dict = {}
            error_flag = False
            for key, value in result.items():
                sub_error, sub_result = _parse_text(repr(value), None, rgx)
                error_flag |= sub_error
                parsed_dict[key] = sub_result
                if error_flag:
                    info(f"parsing error1 for '{key}' value '{value}'")
            return error_flag, parsed_dict

        # Handle lists by recursively parsing each element
        elif isinstance(result, list):
            parsed_list = []
            error_flag = False
            for item in result:
                sub_error, sub_result = _parse_text(repr(item), None, rgx)
                error_flag |= sub_error
                parsed_list.append(sub_result)
                if error_flag:
                    info(f"parsing error2 for value '{item}'")
            return error_flag, parsed_list

        # Handle tuples
        elif isinstance(result, tuple):
            parsed_tuple = []
            error_flag = False
            for item in result:
                sub_error, sub_result = _parse_text(repr(item), None, rgx)
                error_flag |= sub_error
                parsed_tuple.append(sub_result)
            if error_flag:
                info(f"parsing error3 for '{tuple}'")
            return error_flag, tuple(parsed_tuple)

        # Scalar values: Convert to target type if specified
        if target_type:
            type_error, converted = _convert_value(result, target_type)
            if type_error:
                info(f"convert error for '{result}'")
            return type_error, converted

        # Return the scalar value as-is if no target type is specified
        return False, result

    except (ValueError, SyntaxError, TypeError) as e:
        # Handle parsing errors by returning an error flag and None
        info(f"parsing error4 for '{text}'")
        return True, None


def _ast_dictionary(txt):
    """
    Create an AST formatted dictionary string, ensuring proper formatting with quoted keys and values.

    Args:
        txt (str): The input string containing key-value pairs in simplified dictionary-like format.
                   Example: "key1: value1, key2: value2"

    Returns:
        str: An AST formatted dictionary string with single quotes around keys and values.
             Example: "{'key1': 'value1', 'key2': 'value2'}"

    Notes:
        - Ensures the input text is enclosed in curly braces `{}`.
        - Utilizes `format_key_value_pairs` to process individual key-value pairs.
    """
    # Add enclosing braces if not present and content is non-empty
    if txt and (not txt.startswith("{") or not txt.endswith("}")):
        txt = f"{{{txt}}}"

    # Process the content inside the braces and build an AST dictionary string
    text_content = txt.strip("{}")  # Remove braces for processing
    return f"{{{_ast_key_value_pairs(text_content)}}}"


def _ast_key_value_pairs(text):
    """
    Convert a simple string of key-value pairs into an AST-style formatted string.

    This function takes a string containing key-value pairs separated by commas
    (e.g., "key1: value1, key2: value2") and returns a formatted string where
    both the keys and values are enclosed in single quotes
    (e.g., "'key1': 'value1', 'key2': 'value2'").

    Args:
        text (str): A string of key-value pairs, where each pair is separated
                    by a comma and each key and value is separated by a colon.
                    Example: "key1: value1, key2: value2"

    Returns:
        str: A formatted string with keys and values wrapped in single quotes.
             Example: "'key1': 'value1', 'key2': 'value2'"

    Raises:
        ValueError: If a key-value pair does not contain exactly one colon.

    Notes:
        - Leading and trailing whitespace or quotes around keys and values are removed.
        - If a value is empty, it is explicitly set to an empty string (`''`).
        - Each key-value pair is processed independently, and the formatted pairs
          are joined into a single output string.
    """
    formatted_pairs = []  # List to hold formatted key-value pairs
    pairs = text.split(",")  # Split input string into individual key-value pairs

    for pair in pairs:
        key_value = pair.split(":")  # Split each pair into key and value
        if len(key_value) == 2:
            key, value = key_value

            # Clean up the key and value by removing whitespace and surrounding quotes
            key = key.strip().strip("'").strip('"')
            value = value.strip().strip("'").strip('"')

            # Enclose the cleaned key in single quotes
            key = f"'{key}'"

            # Enclose the value in single quotes or set it to an empty string if blank
            value = f"'{value}'" if value else "''"

            # Append the formatted key-value pair to the list
            formatted_pairs.append(f"{key}: {value}")
        else:
            # Raise an error if the pair does not have exactly one colon
            raise ValueError(f"Invalid key-value pair: {pair.strip()}")

    # Combine the formatted pairs into a single string separated by commas
    return ", ".join(formatted_pairs)



def validate_text(text, regex):
    """
    Validates if a string fully matches a specified regex pattern.

    Parameters:
        text (str): The string to validate.
        regex (str): The regex pattern to match.

    Returns:
        bool: True if the string matches the regex or regex is None, False otherwise.
    """
    return bool(re.fullmatch(regex, text)) if text and regex else True


def _convert_value(value, target_type):
    """
    Convert a scalar value to the specified target type.

    Args:
        value: The input scalar value to convert.
        target_type (type): The desired type of the output.

    Returns:
        tuple: (error_flag, result)
            - error_flag (bool): True if conversion fails, False otherwise.
            - result: The converted value, or None if conversion fails.

    Notes:
        - Includes specific logic for converting to `int`, `float`, `bool`, `str`, and `date`.
    """
    try:
        # Convert to integer if possible
        if target_type == int:
            if isinstance(value, int):
                return False, value
            if isinstance(value, str) and value.strip().lstrip('-').isdigit():
                return False, int(value.strip())
            return True, None

        # Convert to float if possible
        elif target_type == float:
            if isinstance(value, (int, float)):
                return False, float(value)
            if isinstance(value, str):
                return False, float(value.strip())
            return True, None

        # Convert to boolean if possible
        elif target_type == bool:
            if isinstance(value, bool):
                return False, value
            if isinstance(value, (int, str)):
                return False, bool(int(value))
            return True, None

        # Convert to string
        elif target_type == str:
            return False, str(value)

        # Convert to date
        elif target_type == date:
            if isinstance(value, str):
                return False, datetime.strptime(value.strip(), "%Y-%m-%d").date()
            return True, None

        # Unsupported target types
        else:
            return True, None

    except (ValueError, TypeError):
        # Handle conversion errors
        return True, None


def data_type(item):
    """
    Identifies the item's data type (List, Dict, int, float, bool, str).

    Parameters:
        item: The item to categorize.

    Returns:
        The identified data type.
    """
    return type(item)


def warn(text):
    pass


def info(text):
    pass



from datetime import date

import pytest

from YMLEditor.structured_text import parse_text, to_text, DEFAULT_FALLBACKS

# Parameters and IDs for positive test cases
positive_test_cases = [  # (text, target_type, test_object)
    ("123", int, 123),  # 1
    ("-23", int, -23),  # 2
    ("3.1415", float, 3.1415),  # 3
    ("True", bool, True),  # 4
    ("False", bool, False),  # 5
    ("New Orleans", str, "New Orleans"),  # 6
    ("{'key7': '42', 'nested': {'bool': 'false', 'list': ['1', 2.5]}}", dict,
     {'key7': '42', 'nested': {'bool': 'false', 'list': ['1', 2.5]}}),  # 7
    ("[85, 'hello', 3.14, True]", list, [85, 'hello', 3.14, True]),  # 8
    ("'2024-09-23'", date, date(2024, 9, 23)),  # 9
    ("0", int, 0),  # 10
    ("3", float, 3),  # 11
    ("-3.1415", float, -3.1415),  # 12

    # Nested structures with mixed types
    ("{'int': 10, 'float': 1.1, 'list': [1, 2, '3']}", dict,
     {'int': 10, 'float': 1.1, 'list': [1, 2, '3']}),  # 13
    ("[{'a': 11}, {'b': 2}]", list, [{'a': 11}, {'b': 2}]),  # 14
    ("{'a': '12'}", dict, {'a': '12'}),  # 15
    # Additional deeply nested structures
    ("{'level1': {'level2': {'level3': 'deep'}}}", dict,
     {'level1': {'level2': {'level3': 'deep'}}}), # 16
    ("[{'nested_dict': {'key': [1, 2, 3]}}, {'another_list': [4.5, 'text']}]", list,
     [{'nested_dict': {'key': [1, 2, 3]}}, {'another_list': [4.5, 'text']}]),  # 17
    ("{'key': [1, {'inner_key': 'value'}, [True, False, '']]}", dict,
     {'key': [1, {'inner_key': 'value'}, [True, False, '']]}),  # 18
    ("[{'a': {'b': {'c': [1, 2, {'d': 3}]}}}, 'end']", list,
     [{'a': {'b': {'c': [1, 2, {'d': 3}]}}}, 'end']),  # 19
    ("{'a': [1, {'b': [2, {'c': '3'}]}]}", dict, {'a': [1, {'b': [2, {'c': '3'}]}]}),  # 20

    # Switches
    ("-s 21", str, "-s 21"),  # 21
]

positive_test_ids = ["1 int:positive", "2 int:negative", "3 float:pi", "4 bool:true",
                     "5 bool:false", "6 str:New_Orleans", "7 dict:nested_structure", "8 list:mixed",
                     "9 date:valid", "10 int:zero", "11 float:no_decimal", "12 float:negative",
                     "13 dict:mixed_types", "14 list:dict_elements", "15 dict:simple",
                     "16 dict:deep_nested_levels", "17 list:deep_nested_mixed",
                     "18 dict:complex_list_structure", "19 list:multiple_nested_dicts",
                     "20 dict:nested_lists_and_dicts", "21 str:Switches"]


# Positive cases
@pytest.mark.parametrize(
    "text, target_type, test_object", positive_test_cases, ids=positive_test_ids, )
def test_to_text(text, target_type, test_object):
    """
    Test to_text for positive cases, including nested and mixed structures.
    """
    print(f"\n\nto_text({test_object},target_type={target_type.__name__})")
    result = to_text(test_object)
    print(f" Expected=: <<{text}>>\n       Got: <<{result}>>")
    assert result == text, f"Failed result: Expected={text}, Got={result}"


# Positive cases
@pytest.mark.parametrize(
    "text, target_type, test_object", positive_test_cases, ids=positive_test_ids, )
def test_parse_text(text, target_type, test_object):
    """
    Test parse_text for positive cases, including nested and mixed structures.
    """
    print(f"\nRunning: text={text}, target_type={target_type}")
    error, result = parse_text(text, target_type)
    print("Result:", result)
    assert result == test_object, f"Failed result: Expected={test_object}, Got={result}"
    assert type(
        result
    ) == target_type, (f"Failed type: Expected={target_type.__name__}, Got="
                       f"{type(result).__name__}")
    assert error == False, f"Failed error_flag: Expected=False, Got={error}"


# Test roundtrip conversion for valid inputs
# Positive cases
@pytest.mark.parametrize(
    "text, target_type, test_object", positive_test_cases, ids=positive_test_ids, )
def test_roundtrip(text, target_type, test_object):
    """
    Test roundtrip: to_text and parse_text back to obj for valid inputs.
    """
    txt = to_text(test_object)
    print(f"\nto_text result: {txt}")
    error_flag, parsed_obj = parse_text(txt, type(test_object))
    assert not error_flag, f"Parsing failed for: {txt}"
    assert parsed_obj == test_object, f"Roundtrip failed: Expected={test_object}, Got={parsed_obj}"


# Test parse_text for negative cases
@pytest.mark.parametrize(
    "text, target_type,  expected_result", [  # Invalid and malformed data
        ("not_a_number1", int, -9999), ("'string2'", int, -9999), ("33.6", int, -9999),
        ("'hello4'", float, -9999.9), ("'2024-05-99'", date, date(1900, 1, 1)),  # 5 Invalid date
        ("{'key6': aaa", dict, None),  # 6 Malformed dict
        ("[7, 2, aaa", list, None),  # 7 Malformed list
        ("3a", int, -9999),  # 8
        ("3b", float, -9999.9),  # 9
    ],
    ids=["1 int:invalid_string", "2 int:from_string", "3 int:float", "4 float:non-numeric_string",
         "5 date:invalid_format", "6 dict:malformed", "7 list:malformed", "8 int:malformed",
         "9 float:malformed", ], )

def test_parse_text_neg(text, target_type, expected_result):
    """
    Test parse_text for negative cases with invalid or malformed inputs.
    """
    print(f"\nRunning: text={text}, target_type={target_type}")
    error, result = parse_text(text, target_type)
    print("Result:", result)
    assert error == True, f"Failed error_flag: Expected=True, Got={error}"
    assert result == expected_result, f"Failed result: Expected={expected_result}, Got={result}"

# Test to_text for unsupported types
@pytest.mark.parametrize(
    "obj, expected_exception", [({1, 2, 3}, TypeError),  # Set is unsupported
                                (complex(1, 2), TypeError),  # Complex numbers are unsupported
                                ], ids=["unsupported:set", "unsupported:complex", ], )
def test_to_text_invalid_types(obj, expected_exception):
    """
    Test to_text for unsupported types.
    """
    with pytest.raises(expected_exception):
        to_text(obj)


@pytest.mark.parametrize(
    "text, target_type, expected_result",
    [("{'empty_list': [], 'empty_dict': {}}", dict, {'empty_list': [], 'empty_dict': {}}), (
            "{'nested': {'level1': {'level2': {'key': 'value'}}}}", dict,
            {'nested': {'level1': {'level2': {'key': 'value'}}}}), ("[]", list, []),
     ("{}", dict, {}), ],
    ids=["dict:empty_structures", "dict:deep_nested", "list:empty", "dict:empty", ], )

def test_parse_text_positive_edge_cases(text, target_type, expected_result):
    """
    Test parse_text for edge cases such as empty and deeply nested structures.
    """
    error, result = parse_text(text, target_type)
    assert result == expected_result, f"Expected={expected_result}, Got={result}"
    assert error == False, f"Expected error flag=False, Got={error}"


@pytest.mark.parametrize(
    "text, target_type, rgx, expected_error_flag, expected_result", [
        ("42", str, r"^\d{2}$", False, "42"),  #  1 two-digit:valid
        ("123", str, r"^\d{2}$", True, "123"), # 2 two_digits:invalid
        ("{'key': 'value', 'number': 42}", dict, r"^\{.*\}$|^\[.*\]$", False,
         {'key': 'value', 'number': 42}),  # 3 dict:valid
        ("[1, 2, 3, 'four']", list, r"^\{.*\}$|^\[.*\]$", False, [1, 2, 3, 'four']),  # 4 list:valid
        ("Not a dict or list", str, r"^\{.*\}$|^\[.*\]$", True, "Not a dict or list"),
        # 5 dict_or_list:invalid
        ("-z 123", str, r"^-z\s+\d+(\s+)?$", False, "-z 123"),  # 6 command_switch:valid
        ("--z 123", str, r"^-z\s+\d+(\s+)?$", True, "--z 123"),# 7 command_switch:invalid_dash
        ("-z", str, r"^-z\s+\d+(\s+)?$", True, "-z"),  # 8 command_switch:missing_value
    ],
    ids=["1 two_digits:valid", "2 two_digits:invalid", "3 dict:valid", "4 list:valid",
            "5 dict_or_list:invalid", "6 command_switch:valid", "7 command_switch:invalid_dash",
            "8 command_switch:missing_value", ]
)

def test_parse_text_with_regex(text, target_type, rgx, expected_error_flag, expected_result):
    """
    Test `parse_text` with regex validation for various cases.
    """
    error_flag, result = parse_text(text, target_type, rgx=rgx)

    # Assertions
    assert error_flag == expected_error_flag, (
        f"Error flag mismatch for text='{text}', rgx='{rgx}'. "
        f"Expected: {expected_error_flag}, Got: {error_flag}")
    assert result == expected_result, (f"Result mismatch for text='{text}', rgx='{rgx}'. "
                                       f"Expected: {expected_result}, Got: {result}")



# yaml_reader.py
import copy
from pathlib import Path
from typing import List, Dict, Any, Type, Optional, Tuple

from cerberus import Validator
from ruamel.yaml import YAML, YAMLError
from ruamel.yaml.comments import CommentedMap, CommentedSeq

def to_plain_dict(data: Any) -> Any:
    """
    Recursively converts ruamel.yaml's CommentedMap/Seq to plain dicts/lists.
    This is essential for compatibility with libraries like Cerberus.
    """
    if isinstance(data, CommentedMap):
        return {key: to_plain_dict(value) for key, value in data.items()}
    if isinstance(data, CommentedSeq):
        return [to_plain_dict(item) for item in data]
    return data


class ConfigLoader:
    def __init__(
            self, schema: Dict = None,
            error_modifications: Optional[List[Dict]] = None,
            validator_class: Type[Validator] = Validator
    ):
        """Initializes the ConfigLoader."""
        self.schema = schema
        self.validator_class = validator_class
        self.error_modifications = error_modifications or ERROR_MODIFICATION_RULES
        self.ruamel_parser = YAML(typ='safe')

    def read(
            self,
            config_file: Path,
            allow_unknown: bool = False,
            root_node: str | None = None,
            block_descriptor: str | None = None,
            normalize: bool = False
    ) -> dict:
        """Loads and validates a YAML configuration file against the schema."""
        if not isinstance(config_file, Path):
            config_file = Path(config_file)
        if not config_file.exists():
            raise FileNotFoundError(f"Configuration file not found: {Path.cwd()}/{config_file}")

        yaml_text = config_file.read_text(encoding="utf-8")

        validated_data, line, error_text = self._validate_text(
            yaml_text, normalize=normalize, allow_unknown=allow_unknown
        )

        if error_text:
            error_location = f"on or near line {line} " if line else ""
            error_message = (
                f"  ❌ Error: Configuration file '{config_file}' has an error {error_location}:\n\n"
                f"    {error_text}"
            )
            raise ValueError(error_message)

        return validated_data

    def validate_text(
            self,
            yaml_text: str,
            normalize: bool = False,
            allow_unknown: bool = False
    ) -> Tuple[Optional[Dict], Optional[int], Optional[str]]:
        """Public method to validate a string of YAML text for GUI use."""
        return self._validate_text(yaml_text, normalize=normalize, allow_unknown=allow_unknown)

    def _validate_text(
            self,
            yaml_text: str,
            normalize: bool = False,
            allow_unknown: bool = False
    ) -> Tuple[Optional[Dict], Optional[int], Optional[str]]:
        """
        Internal method that contains the core validation logic for a YAML string.
        """
        try:
            # Step 1: Load with ruamel to get line numbers and comment-aware data
            ruamel_data = self.ruamel_parser.load(yaml_text) or {}
        except YAMLError as e:
            line = e.problem_mark.line + 1 if hasattr(e, 'problem_mark') else None
            problem, suggestion = self._translate_error_message(str(e))
            error_message = f"Problem: YAML syntax error.\n    {problem}\n    Suggestion: {suggestion}"
            return None, line, error_message

        if self.schema:
            # --- THE DEFINITIVE FIX ---
            # Step 2: Convert the ruamel_data to a plain dict for Cerberus
            plain_data_for_validation = to_plain_dict(ruamel_data)

            # Step 3: Validate the plain dict using a fresh validator and a copied schema
            schema_copy = copy.deepcopy(self.schema)
            validator = self.validator_class(schema_copy)
            validator.allow_unknown = allow_unknown

            if not validator.validate(plain_data_for_validation):
                field, issues = next(iter(validator.errors.items()))
                path_list = field.split('.')
                issue = issues[0]

                # Step 4: Use the original ruamel_data to find the line number
                error_node = self._get_node_by_path(ruamel_data, path_list)
                line = error_node.lc.line + 1 if hasattr(error_node, 'lc') else None

                problem, suggestion = self._translate_cerberus_issue(validator, issue, path_list)
                found_value = self._get_node_by_path(ruamel_data, path_list)
                error_message = (
                    f"Problem: {problem}\n"
                    f"    Found Value: {repr(found_value)}\n"
                    f"    Suggestion: {suggestion}"
                )
                return None, line, error_message

            # Return the plain, normalized data on success
            return validator.normalized(plain_data_for_validation) if normalize else plain_data_for_validation, None, None

        # If no schema, return the plain data
        return to_plain_dict(ruamel_data), None, None

    # ... (the rest of the file is unchanged) ...
    def _translate_error_message(self, error_string: str) -> tuple[str, str]:
        for rule in self.error_modifications:
            trigger_phrase = rule.get("trigger_phrase", "")
            if trigger_phrase and trigger_phrase in error_string:
                return rule["problem"], rule["suggestion"]
        return f"An unknown error occurred: {error_string}", "Please check file syntax."

    def _translate_cerberus_issue(self, validator: Validator, issue: str, path: List[str]) -> tuple[str, str]:
        field_name = path[-1] if path else "a setting"
        problem, suggestion = self._translate_error_message(issue)
        problem = problem.format(field_name=field_name)
        suggestion = suggestion.format(field_name=field_name)

        if "unallowed value" in issue:
            schema_rule = self._get_schema_rule_from_path(validator.schema, path)
            if schema_rule and 'allowed' in schema_rule:
                allowed_values = ", ".join(f"'{v}'" for v in schema_rule['allowed'])
                suggestion += f" Allowed values are: {allowed_values}."
        return problem, suggestion

    def _get_schema_rule_from_path(self, schema: Any, path: List[str]) -> Optional[Dict]:
        current_schema_item = schema
        for key in path:
            if not isinstance(current_schema_item, dict): return None
            current_schema_item = current_schema_item.get(key)
            if current_schema_item is None: return None
            if isinstance(current_schema_item, dict) and 'schema' in current_schema_item:
                current_schema_item = current_schema_item['schema']
        return current_schema_item if isinstance(current_schema_item, dict) else None

    @staticmethod
    def _get_node_by_path(data: Any, path: List[str]) -> Any:
        current = data
        for key in path:
            try:
                if isinstance(current, list) and key.isdigit():
                    current = current[int(key)]
                elif hasattr(current, 'get'):
                    current = current.get(key)
                else:
                    return None
            except (KeyError, IndexError, TypeError):
                return None
        return current


ERROR_MODIFICATION_RULES = [
    # --- Rules for modifying  ruamel.yaml error text ---
    {
        "key": "YAML_MAPPING_ERROR",
        "trigger_phrase": "expected '<block end>', but found '<block mapping start>'",
        "problem": "Indentation or colon error.",
        "suggestion": "Check for a missing colon (:) after a key or incorrect indentation for the items below it."
    },
    # ... other ruamel.yaml error rules ...

    # --- Rules for modifying Cerberus  error text ---
    {
        "key": "REQUIRED_FIELD",
        "trigger_phrase": "required field",
        "problem": "The required setting '{field_name}' is missing.",
        "suggestion": "Please add '{field_name}: <value>' to this section.",
    },
    {
        "key": "UNKNOWN_FIELD",
        "trigger_phrase": "unknown field",
        "problem": "The setting '{field_name}' is not recognized here.",
        "suggestion": "Please remove it or check for a typo.",
    },
    {
        "key": "UNALLOWED_VALUE",
        "trigger_phrase": "unallowed value",
        "problem": "The value for '{field_name}' is not allowed.",
        "suggestion": "Please choose one of the permitted values."
    },
    {
        "key": "NULL_VALUE",
        "trigger_phrase": "null value not allowed",
        "problem": "The setting '{field_name}' cannot be empty.",
        "suggestion": "Please provide a value for '{field_name}'."
    }
]


import pytest

from YMLEditor.data_manager import DataHandler


class TestDataHandler:
    """Tests for the DataHandler class."""

    class ConcreteDataHandler(DataHandler):
        """Concrete implementation of DataHandler for testing."""

        def get(self, data, key):
            return super().get(data, key)

        def set(self, data, key, value):
            super().set(data, key, value)

        def items(self, data):
            return super().items(data)

    @pytest.fixture
    def handler(self):
        """Fixture to provide a ConcreteDataHandler instance."""
        return self.ConcreteDataHandler(verbose=0)

    def test_insert_valid(self, handler):
        data = [1, 2, 3]
        handler.insert(data, 1, 99)
        assert data == [1, 99, 2, 3]

    def test_insert_invalid_index(self, handler):
        data = [1, 2, 3]
        with pytest.raises(IndexError):
            handler.insert(data, 5, 99)

    def test_insert_non_list(self, handler):
        data = {"a": 1}
        with pytest.raises(ValueError):
            handler.insert(data, 0, 99)

    def test_delete_dict_key_exists(self, handler):
        data = {"a": 1, "b": 2}
        handler.delete(data, "a")
        assert data == {"b": 2}

    def test_delete_dict_key_missing(self, handler):
        data = {"a": 1}
        with pytest.raises(KeyError):
            handler.delete(data, "b")

    def test_delete_list_index_valid(self, handler):
        data = [1, 2, 3]
        handler.delete(data, 1)
        assert data == [1, 3]

    def test_delete_list_index_invalid(self, handler):
        data = [1, 2, 3]
        with pytest.raises(IndexError):
            handler.delete(data, 5)

    def test_delete_invalid_type(self, handler):
        data = 123
        with pytest.raises(TypeError):
            handler.delete(data, 1)

    def test_get_nested_key(self, handler):
        data = {"a": {"b": {"c": 42}}}
        assert handler.__getitem__(data, "a.b.c") == 42

    def test_get_missing_key(self, handler):
        data = {"a": {"b": {"c": 42}}}
        assert handler.__getitem__(data, "a.x.c") is None

    def test_set_nested_key(self, handler):
        data = {"a": {"b": {}}}
        handler.__setitem__(data, "a.b.c", 99)
        assert data == {"a": {"b": {"c": 99}}}

    def test_set_create_missing(self, handler):
        data = {"a": {}}
        handler.__setitem__(data, "a.b.c", 99)
        assert data == {"a": {"b": {"c": 99}}}

    def test_items_dict(self, handler):
        data = {"a": 1, "b": 2}
        items = list(handler.items(data))
        assert items == [("a", 1), ("b", 2)]

    def test_items_list(self, handler):
        data = [10, 20, 30]
        items = list(handler.items(data))
        assert items == [(0, 10), (1, 20), (2, 30)]

    def test_items_invalid_type(self, handler):
        data = 123
        with pytest.raises(TypeError):
            list(handler.items(data))

    def test_replace_indirect_valid(self, handler):
        data = {"INDIRECT": "_suffix", "KEY_suffix": 99}
        resolved_key = handler.replace_indirect(data, "KEY@INDIRECT")
        assert resolved_key == "KEY_suffix"

    def test_replace_indirect_invalid(self, handler):
        data = {"INDIRECT": "_suffix"}
        resolved_key = handler.replace_indirect(data, "KEY@MISSING")
        assert resolved_key is None

    def test_replace_indirect_no_at_symbol(self, handler):
        data = {"INDIRECT": "_suffix"}
        resolved_key = handler.replace_indirect(data, "KEY")
        assert resolved_key == "KEY"


class TestNavigateHierarchy:
    """Test suite for the _navigate_hierarchy method."""

    @pytest.fixture
    def mock_manager(self):
        """Mock DataManager with _navigate_hierarchy and _validate_index methods."""
        class MockDataManager:
            def _navigate_hierarchy(self, *args, **kwargs):
                return DataHandler._navigate_hierarchy(self, *args, **kwargs)
            def _validate_index(self, key):
                return DataHandler._validate_index(self, key)

        return MockDataManager()

    def test_navigate_dict_success(self, mock_manager):
        data = {"a": {"b": {"c": 1}}}
        target, key = mock_manager._navigate_hierarchy(data, "a.b.c", create_missing=False)
        assert target == {"c": 1}
        assert key == "c"

    def test_navigate_list_success(self, mock_manager):
        data = [{"x": 1}, {"y": 2}]
        target, key = mock_manager._navigate_hierarchy(data, "1.y", create_missing=False)
        assert target == {"y": 2}
        assert key == "y"

    def test_create_missing_dict(self, mock_manager):
        data = {"a": {"b": {}}}
        target, key = mock_manager._navigate_hierarchy(data, "a.b.c", create_missing=True)
        assert target == {}
        assert key == "c"

    def test_create_missing_list(self, mock_manager):
        data = []
        target, key = mock_manager._navigate_hierarchy(data, "2.x", create_missing=True)
        assert len(data) == 3  # List was extended
        assert target == {}
        assert key == "x"

    def test_index_out_of_range(self, mock_manager):
        data = [{"x": 1}]
        with pytest.raises(IndexError):
            mock_manager._navigate_hierarchy(data, "2.x", create_missing=False)

    def test_invalid_index(self, mock_manager):
        data = [{"x": 1}]
        with pytest.raises(ValueError):
            mock_manager._navigate_hierarchy(data, "a.x", create_missing=False)

    def test_type_error(self, mock_manager):
        data = 42  # Invalid type
        with pytest.raises(TypeError):
            mock_manager._navigate_hierarchy(data, "a.b.c", create_missing=True)




#  Copyright (c) 2024.
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the “Software”), to deal in the
#   Software without restriction,
#   including without limitation the rights to use, copy, modify, merge, publish, distribute,
#   sublicense, and/or sell copies
#   of the Software, and to permit persons to whom the Software is furnished to do so, subject to
#   the following conditions:
#  #
#   The above copyright notice and this permission notice shall be included in all copies or
#   substantial portions of the Software.
#  #
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#   BUT NOT LIMITED TO THE
#   WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
#   EVENT SHALL THE AUTHORS OR
#   COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
#   CONTRACT, TORT OR
#   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#  #
#   This uses QT for some components which has the primary open-source license is the GNU Lesser
#   General Public License v. 3 (“LGPL”).
#   With the LGPL license option, you can use the essential libraries and some add-on libraries
#   of Qt.
#   See https://www.qt.io/licensing/open-source-lgpl-obligations for QT details.

#
#
from typing import List

# Use PySide for imports and fall back to PyQt
try:
    from PySide6.QtWidgets import QWidget, QLabel, QGridLayout, QSpacerItem, QSizePolicy, \
        QVBoxLayout
except ImportError:
    from PyQt6.QtWidgets import QWidget, QLabel, QGridLayout, QSpacerItem, QSizePolicy, QVBoxLayout

from YMLEditor.item_widget import ItemWidget


class SettingsWidget(QWidget):
    """
    Provides a configurable user interface for editing settings in a config file.

    - Displays settings in a vertical layout based on the supplied row definitions which include:
      type of control (text field, combo box, etc.) and the key for retrieving
      each field. See project readme for a list of supported widgets and options.
    - Changes made to widgets are synched back to the `Config` data.
    - `Config` must support `get`and`set` data and `load` and `save` data to a config file.
    - Validates user input according to provided rules (regular expressions).
    - Highlights the entry if there is an error.
    - Supports callback when specified keys change.
    - Supports data fields that are lists, dictionaries, or scalar (int, string, etc.)
    - Does not support fields that are complex nested data structures.
    - Supports switching between multiple formats (e.g., "basic" vs "expert" format).

    **Methods**:
    """

    def __init__(
            self, config, formats, mode, trigger_keys=None, callback=None, verbose=3,
            text_edit_height=60, error_style="color: orange;"
            ):
        """
        Init

        Args:
            config (Config): Configuration object to load, update, and store settings.
            formats (dict): Display formats for different modes.
            mode (str): Used to select between multiple layout formats.
            trigger_keys (List[str], optional): Keys that trigger a callback or redisplay. Defaults to None.
            callback (Callable[[str, Any], None], optional): Function called when user updates any trigger key. Defaults to None.
            verbose (int, optional): Verbosity level. Defaults to 3 (warning).
            text_edit_height (int, optional): Height for text-edit fields. Defaults to 60.
            error_style (str, optional): CSS style for validation errors. Defaults to "color: orange;".
        """
        super().__init__()

        self.config = config
        self.verbose = verbose
        self.formats = formats
        self.error_style = error_style
        self._mode = mode  # Select which format within formats to use
        self._validate_format(formats, mode)
        self.ignore_changes = False
        self.is_loaded = False
        self.trigger_keys = trigger_keys  # Keys that trigger callback
        self.config_widgets = []
        self.text_edit_height = text_edit_height
        self.callback = callback

        # Set up the top-level layout once
        self.main_layout = QVBoxLayout(self)
        self.grid_layout = QGridLayout()
        self.main_layout.addLayout(self.grid_layout)
        self._setup_ui()

    def _setup_ui(self):
        """
        Create and arrange widgets based on the format.

        Raises:
            ValueError: If 'self.formats' is not set or is not a dictionary.
            KeyError: If the key in 'self.mode' is not found in 'self.formats'.
            Exception: If an error occurs while setting up a widget, with details on the row and
            key.
        """
        # Clear the grid layout without removing the top-level layout
        self.clear_layout()

        row, data_key, format_row = -1, None, None
        fmt = self.formats[self._mode]  # Get format for the current mode

        try:
            for row, (data_key, format_row) in enumerate(fmt.items()):
                # Unpack format_row with support for an optional "style" element
                if len(format_row) == 4:
                    label_text, widget_type, options, width = format_row
                    style = None
                elif len(format_row) == 5:
                    # Optional style element
                    label_text, widget_type, options, width, style = format_row

                # Create specified Widget
                config_item = ItemWidget(
                    self.config, widget_type, None, options, self.on_change, width, data_key,
                    verbose=self.verbose, text_edit_height=self.text_edit_height,
                    error_style=self.error_style, style=style
                )

                self.config_widgets.append(config_item)

                # Add label to col 0
                label = QLabel(label_text)
                label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
                self.grid_layout.addWidget(label, row, 0)

                # Add widget to col 1
                self.grid_layout.addWidget(config_item.widget, row, 1)

                # Add spacer item to col 2 to force items left
                h_spacer = QSpacerItem(
                    1, 1, QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum
                )
                self.grid_layout.addItem(h_spacer, row, 2)
        except Exception as e:
            raise Exception(
                f"Error setting up widget at row {row} with key '{data_key}'\n"
                f"{format_row}'.\n"
                f"{str(e)}"
            ) from e

        # Add an expanding spacer item
        v_spacer = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.grid_layout.addItem(v_spacer, self.grid_layout.rowCount(), 0, 1, 3)


    def set_layout(self, mode):
        """
        Change the current display mode and redisplay the UI.

        Args:
            mode (str): The new mode to switch to.

        Raises:
            ValueError: If the specified mode is not a valid key in `formats`.
        """
        if mode not in self.formats:
            raise ValueError(f"Invalid mode '{mode}'. Must be one of: {list(self.formats.keys())}")

        self._mode = mode
        self.format = self.formats[mode]  # Update the format based on the new mode

        self.clear_layout()
        self._setup_ui()
        self.display()

    def display(self):
        """
        Update the display widgets with latest data from Config. Each widget's value is
        set based on the corresponding configuration key's value.
        """
        self.ignore_changes = True  # Temporarily ignore changes during synchronization
        if len(self.config_widgets) == 0:
            self._setup_ui()  # Initialize the UI if it hasn't been set up

        # Iterate over each item in the layout to update widget values from config
        for widget_item in self.config_widgets:
            widget_item.display()

        self.ignore_changes = False  # Re-enable change tracking after sync
        self.is_loaded = True  # Mark the display as loaded

    def save(self):
        """
        Save the current data to the configuration file
        """
        if self.is_loaded:
            self.config.save()

    def undo(self):
        """
        Reverts data to last snapshot and refreshes display
        """
        self.config.snapshot_undo()
        self.display()

    def on_change(self, key, value):
        """
        Called by ItemWidget after it has updated data from a user edit if
        the key is in `trigger_keys`.

        Args:
            key (str): The key that changed.
            value (str): The value that changed.
        """
        # Call callback if key is in trigger_keys.
        if key and self.trigger_keys is not None:
            if key in self.trigger_keys:
                if self.callback:
                    self.callback(key, value)
                else:
                    self.display()


    def clear_layout(self):
        """
        Clear the current settings widget layout.
        """
        self._clear_layout(self.grid_layout)

    def _clear_layout(self, layout):
        """
        Remove all items from a layout.

        Args:
            layout (QLayout): The layout to clear.
        """
        self.config_widgets = []
        self.is_loaded = False
        while layout.count():
            item = layout.takeAt(0)

            widget = item.widget()
            if widget:
                widget.deleteLater()

            nested_layout = item.layout()
            if nested_layout:
                self._clear_layout(nested_layout)

    def _validate_format(self, formats, mode):
        """
        Validate the display format

        Args:
            formats (dict): A dictionary defining display formats.
            mode (str): Used to select between multiple layout formats

        Raises:
            ValueError: If 'formats' is not formatted properly.
            KeyError: If 'mode' is not a format in formats.
        """
        # Validate 'formats' as a non-empty dictionary
        if not formats or not isinstance(formats, dict):
            raise ValueError(
                f"'formats' must be a non-empty dictionary. Current type: {type(formats).__name__}"
            )

        # Validate format mode is in 'formats'
        if mode not in formats:
            raise KeyError(
                f"Unknown mode '{mode}' in 'formats'. Available modes: {list(formats.keys())}"
            )

        format_def = formats[mode]  # Get format for the current mode

        # Validate the entries in the format
        for row, (key, value) in enumerate(format_def.items()):
            # Validate format line
            if not isinstance(value, tuple):
                raise ValueError(
                    f"Format for key '{key}' is invalid at row {row}. Expected tuple:  "
                    f"(label_text, widget_type, options, width), but got: {value}"
                )



#  Copyright (c) 2024.
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the “Software”), to deal in the
#   Software without restriction,
#   including without limitation the rights to use, copy, modify, merge, publish, distribute,
#   sublicense, and/or sell copies
#   of the Software, and to permit persons to whom the Software is furnished to do so, subject to
#   the following conditions:
#  #
#   The above copyright notice and this permission notice shall be included in all copies or
#   substantial portions of the Software.
#  #
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#   BUT NOT LIMITED TO THE
#   WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
#   EVENT SHALL THE AUTHORS OR
#   COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
#   CONTRACT, TORT OR
#   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#
from abc import ABC, abstractmethod
import copy
import shutil
from datetime import datetime
import os
import fnmatch
from typing import List, Any, Dict, Iterator, Union


class DataManager(ABC):
    """
       Base class for managing data, providing functions for: load, save, get, set, and
       undo.

       Subclasses are created for specific file formats and must support: `_load_data` and
       `_save_data`.

       Attributes:
           _data (Union[Dict, List]): The main data structure being managed (dictionary or list).
           file_path (str): Path to the file associated with the data.
           directory (str): Directory containing the file.
           unsaved_changes (bool): Indicates if there are unsaved changes in the data.
           snapshots (List): Stack of snapshots for supporting undo functionality.
           max_snapshots (int): Maximum number of snapshots retained for undo operations.
           proxy_mapping (Dict): Maps keys to proxy files for granular build system dependencies.

        **Methods**:
       """

    def __init__(self, verbose=3, archive_data=False):
        """
        Init

        Args:
            verbose (int): Verbosity level for logging warnings. Defaults to 1.
            archive_data (bool): If true, data copies will be stored.

        Attributes:
            verbose (int): Verbosity level for logging and debugging.
            _data: Internal data structure being managed and manipulated.
            file_path (str | None): Path to the file associated with the current data.
            directory (str | None): Directory associated with the current data or file
                operation.
            unsaved_changes (bool): Flag to determine whether the data has unsaved
                modifications.
            snapshots (list): Stack of snapshots representing the stored states of the
                `_data` attribute for undo functionality.
            handler (DataHandler): Handler for the specific data type, used for
                managing data operations and interactions.
            max_snapshots (int): Maximum number of snapshots to retain within the
                `snapshots` stack.
            proxy_mapping (dict): Dictionary mapping keys to proxy files for certain
                data operations or file handling.
            error (str): String representation of the last error or issue encountered
                during an operation.
        """
        self.verbose = verbose
        self.archive_data = archive_data
        self._data = None
        self.file_path = None
        self.directory = None
        self.unsaved_changes = False
        self.snapshots = []  # stack to store snapshots of _data for undo
        self.handler = DataHandler(verbose)  # the handler for our data type
        self.max_snapshots = 20  # Maximum number of snapshots to retain for undo
        self.proxy_mapping = {}  # Dictionary to map keys to proxy files
        self.error = ""

    @abstractmethod
    def _load_data(self, f) -> Union[Dict, List]:
        """
         Load data from a file and return as a dictionary or list.
         Subclasses must implement this method to define file-specific behavior.

         Args:
             f: File object to load data from.

         Returns:
             Union[Dict, List]: The loaded data.

         Raises:
            ValueError: If the file is empty or cannot be parsed.
         """
        raise NotImplementedError(
            f"Subclass must implement {self.__class__.__name__}._load_data()"
        )

    @abstractmethod
    def _save_data(self, f, data):
        """
        Save data to a file.
        Subclasses must implement this method to define file-specific behavior.

        Args:
            f: File object to save data to.
            data: Data to be saved.
        Raises:
            RuntimeError: If the file cannot be saved.
        """
        raise NotImplementedError(
            f"Subclass must implement {self.__class__.__name__}._save_data()"
        )

    def load(self, path):
        """
        Load data from the specified file and initialize state.
        Create an initial snapshot for undo functionality.

        Args:
            path (str): Path to the file to load.

        Returns:
            bool: True if the file was loaded successfully.

        Raises:
            FileNotFoundError: If the file does not exist.
            IOError: If there is an issue reading the file.
            ValueError: If the file contents are invalid.
        """
        self.error = ""
        # Clear any proxy keys
        self.proxy_mapping = {}  # Dictionary to map keys to proxy files
        self.file_path = path
        self.directory = os.path.dirname(path)

        try:
            with open(path, self.get_open_mode()) as f:
                self.init_data(self._load_data(f))
                self.unsaved_changes = False
            return True
        except FileNotFoundError:
            self.error = f"Error: File not found: {path}"
        except IOError as e:
            self.error = f"Error: File: {path}\n{e}"
        except ValueError as e:
            self.error = f"Error: Invalid file contents for {self.__class__.__name__}: {path}\n{e}"
        except Exception as e:
            self.error = f"Error loading:  {self.__class__.__name__}: {path}\n{e}"
        return False

    def get_open_mode(self, write=False):
        """
        Provides the file mode for the data file.
        Default is 'w' for write mode, 'r' for read mode.
        Override for binary or other modes.

        Args:
            write (bool): Whether the file is being opened for writing.

        Returns:
            str: 'w' for write mode, 'r' for read mode.
        """
        return 'w' if write else 'r'

    def save(self):
        """
        Save the current data to the file if it has been modified.
        Create an in-memory snapshot of the data for undo feature.

        Returns:
            bool: True if the file was saved successfully.

        Raises:
            ValueError: If the file path or data is None.
        """
        self.error = ""
        if self.file_path is None:
            raise ValueError("Save Error: File path: cannot be None")

        if self._data is None:
            raise ValueError("Save Error: Data cannot be None")

        if self.unsaved_changes:
            # For the first save (snapshot is empty) also create an archive file copy
            if len(self.snapshots) == 1 and self.archive_data:
                self.create_archive(self.file_path, "snapshots")

            self.snapshot_push()  # Push the current data to snapshot stack
            try:
                with open(self.file_path, self.get_open_mode(write=True)) as f:
                    self._save_data(f, self._data)
                self.unsaved_changes = False
                return True
            except Exception as e:
                self.error = f"Error saving: {self.file_path}\n{e}"
                return False

    def get(self, key_or_index, default=None):
        """
        Retrieve a value from the data, returning a default if not found.

        Args:
            key_or_index: Key or index for the data.
            default: Default value to return if the key or index is not found.

        Returns:
            The value associated with the key or index, or the default.
        """
        value = self.__getitem__(key_or_index)
        return value if value is not None else default

    def create_archive(self, file_path, sub_directory=""):
        """
        Creates a timestamped archive of a given file in the same directory or in a specified sub-directory.

        Args:
            file_path (str): Path to the file to snapshot.
            sub_directory (str): Optional subdirectory to save the archive in.

        Returns:
            str: Path to the snapshot file if successful, None otherwise.
        """
        if not os.path.isfile(file_path):
            return None

        base_dir = os.path.dirname(file_path)
        file_name = os.path.basename(file_path)
        name, ext = os.path.splitext(file_name)
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M")

        # Build the snapshot path
        if sub_directory:
            snapshot_dir = os.path.join(base_dir, sub_directory)
            os.makedirs(snapshot_dir, exist_ok=True)
        else:
            snapshot_dir = base_dir

        snapshot_path = os.path.join(snapshot_dir, f"{name}_{timestamp}{ext}")

        try:
            shutil.copy2(file_path, snapshot_path)
            return snapshot_path
        except IOError as e:
            print(f"Error creating snapshot: {e}")
            return None

    def __getitem__(self, key_or_index):
        """
        Retrieve a value from the data using the specified key or index.

        Args:
            key_or_index: Key or index for the data.

        Returns:
            The value associated with the specified key or index.
        """
        return self.handler.__getitem__(self._data, key_or_index)

    def __setitem__(self, key, value):
        """
        Update the data with a new value at the specified key or index.

        Args:
            key: Key or index for the data.
            value: New value to set.
        """
        self.handler.__setitem__(self._data, key, value)
        self.unsaved_changes = True

    def insert(self, key, value):
        """
        Insert a new item into the data.

        Args:
            key: Key or index for the new item.
            value: Value to insert.
        """
        self.handler.insert(self._data, key, value)
        self.unsaved_changes = True

    def delete(self, key):
        """
        Remove an item from the data.

        Args:
            key: Key or index of the item to remove.
        """
        self.handler.delete(self._data, key)
        self.unsaved_changes = True

    def snapshot_undo(self):
        """
        Restore the data to the previous state using the snapshot stack.

        The first snapshot remains in the stack to always allow undo to initial state.
        """
        if not self.snapshots:
            return

        # Pop last snapshot unless it is the only one left, always keep initial snapshot
        if len(self.snapshots) > 1:
            self._data = self.snapshots.pop()
        else:
            self._data = copy.deepcopy(self.snapshots[0])

        self.unsaved_changes = True  # Data has been modified

    def snapshot_push(self):
        """
        Push the current state of the data to snapshot stack for undo functionality.

        If the maximum number of snapshots is reached, the second-oldest snapshot is removed.
        The oldest is always retained for return to initial state.
        """
        if len(self.snapshots) >= self.max_snapshots:
            # Stack is full.  Remove the second-oldest snapshot
            self.snapshots.pop(1)

        # Create a deep copy of _data to store as a snapshot
        self.snapshots.append(copy.deepcopy(self._data))

    def __len__(self):
        """
        Get the number of items in the data.

        Returns:
            int: Number of items in the data.
        """
        if self._data:
            return len(self._data)
        else:
            return 0

    def items(self):
        """
        Get an iterator over the items in the data.

        Returns:
            Iterator: An iterator over the data items.
        """
        return self.handler.items(self._data)

    def init_data(self, data: Union[Dict[str, Any], List[Any]]):
        """
        Initialize the data

        Args:
            data (Dict[str, Any]): New data to initialize.
        """
        self.snapshots = []
        self._data = data
        self.snapshot_push()  # Save the initial state
        self.unsaved_changes = True

    def create(self, data: Dict[str, Any]):
        """
        Create a new config file with the specified data.

        Args:
            data (Dict[str, Any]): Data to save in the new file.
        """
        self.init_data(data)
        self.save()

    def set(self, key, value):
        """
        Update the data at key with the new value and touch proxy file if key in proxy list.

        Args:
            key: Key for the data.
            value: New value to set.
        """
        self.unsaved_changes = True
        self.__setitem__(key, value)

        # Check if updating this key should trigger a touch to a proxy_file

        # See if this key is in the proxy_mapping
        if proxy_file := self._get_proxy(key):
            touch_file(proxy_file)

    def register_proxy_file(self, proxy_file, update_keys):
        """
        Registers a proxy file and associates it with specific configuration keys.

        A proxy file is updated ("touched") when any of its associated keys are modified.
        This helps manage build dependencies by ensuring that only relevant changes
        in the configuration trigger a rebuild of dependent components.

        Args:
            proxy_file (str): Path to the proxy file that will be updated.
            update_keys (List[str]): Keys in the configuration that, when updated,
                                     trigger the proxy file to be touched.  Support wildcard keys such as A.*

        Raises:
            ValueError: If any of the keys in `update_keys` are already linked to another proxy file.
        """
        for key in update_keys:
            if key in self.proxy_mapping:
                raise ValueError(
                    f"Key '{key}' is already associated with another proxy file: "
                    f"{self.proxy_mapping[key]}"
                )
            self.proxy_mapping[key] = proxy_file

    def _get_proxy(self, key):
        """
        Retrieve the proxy file associated with a given key, including wildcard-based matches.

        Wildcards supported:
            - A.* will match A.xyz, A.color, A.123 etc.

        Args:
            key (str): The key to look up.

        Returns:
            str or None: The proxy file associated with the key, or None if not found.
        """
        # Fill in value for keys with indirect fields (e.g. "@XYZ")
        key = self.handler.replace_indirect(self._data, key)

        # Try direct match first
        if key in self.proxy_mapping:
            self.debug(f"Found proxy file for key '{key}'")
            return self.proxy_mapping[key]

        # Fallback: match against wildcard entries
        for pattern, proxy_file in self.proxy_mapping.items():
            if "*" in pattern and fnmatch.fnmatch(key, pattern):
                self.debug(f"Found proxy wildcard pattern '{pattern}' for key '{key}'")
                return proxy_file

        return None

    def warn(self, text):
        if self.verbose > 2:
            print(f"Warning: {text}")

    def info(self, text):
        if self.verbose > 3:
            print(f"Info: {text}")

    def debug(self, text):
        if self.verbose > 4:
            print(f"Debug: {text}")


def touch_file(filename):
    """
    Set the file's modification and access time to the current time.

    Args:
        filename (str): Path to the file.
    """
    with open(filename, 'a'):
        os.utime(filename, None)


class DataHandler(ABC):
    """
    A class for managing nested dictionary and list structures with abstract methods
    for getting, setting, and iterating over elements.
    """

    def __init__(self, verbose: int = 3):
        """
        Init

        Args:
            verbose (int): Verbosity level for logging warnings. Defaults to 3 (WARNING).
        """
        self.verbose = verbose

    def insert(self, data_list: list, idx: int, item: Any) -> None:
        """
        Insert an item into a list at the specified index.

        Args:
            data_list (list): The list to modify.
            idx (int): The index to insert the item.
            item (Any): The item to insert.

        Raises:
            ValueError: If the data is not a list.
            IndexError: If the index is out of range.
        """
        if not isinstance(data_list, list):
            raise ValueError("Insert operation is only supported for lists.")
        if not (0 <= idx <= len(data_list)):
            raise IndexError(f"Index {idx} is out of range for insertion.")
        data_list.insert(idx, item)

    def delete(self, data: Union[Dict, list], idx: Union[str, int]) -> None:
        """
        Delete an item from a dictionary or list.

        Args:
            data (Dict or list): The data structure to modify.
            idx (str or int): The key (for dict) or index (for list) to delete.

        Raises:
            TypeError: If the data type is unsupported.
            KeyError: If the key is not found in a dictionary.
            IndexError: If the index is out of range for a list.
        """
        if isinstance(data, dict):
            if idx in data:
                del data[idx]
            else:
                raise KeyError(f"Key '{idx}' not found in dictionary.")
        elif isinstance(data, list):
            if not isinstance(idx, int):
                raise TypeError(f"List indices must be integers, got '{type(idx).__name__}'.")
            if 0 <= idx < len(data):
                del data[idx]
            else:
                raise IndexError(f"Index {idx} out of range for list.")
        else:
            raise TypeError("Delete operation is only supported for dicts and lists.")

    def __getitem__(self, data: Union[Dict, list], key) -> Any:
        """Retrieve a value from nested data."""
        return self._access_item(data, key, set_item=False)

    def __setitem__(self, data: Union[Dict, list], key, value: Any) -> None:
        """Set a value in nested data."""
        self._access_item(data, key, value=value, set_item=True)

    def _access_item(
            self, data: Union[Dict, list], key: str, value: Any = None, set_item: bool = False
    ) -> Any:
        """
        Internal method for getting or setting values in nested data structures.
        """
        key = self.replace_indirect(data, key)

        try:
            container, final_key = self._navigate_hierarchy(data, key, create_missing=set_item)
            if isinstance(container, dict):
                if set_item:
                    container[final_key] = value
                else:
                    return container[final_key]
            elif isinstance(container, list):
                index = self._validate_index(final_key)
                if set_item:
                    while index >= len(container):
                        container.append(None)
                    container[index] = value
                else:
                    return container[index]
            else:
                raise TypeError("Unsupported container type.")
        except (KeyError, IndexError, ValueError, TypeError) as e:
            return None

    def _navigate_hierarchy(self, data: Union[Dict, List], key: str, create_missing: bool = False):
        """
        Navigate nested hierarchies within dictionaries or lists.

        Args:
            data (Union[Dict, List]): The data structure to navigate.
            key (str): A dot-separated key path (e.g., "a.b.2.c").
            create_missing (bool): Whether to create missing intermediate nodes.

        Returns:
            Tuple[Union[Dict, List], str]: A tuple containing the target container and the last
            key or index.

        Raises:
            KeyError: If a dictionary key is missing and `create_missing` is False.
            IndexError: If a list index is out of range and `create_missing` is False.
            TypeError: If an unsupported type is encountered during navigation.
            ValueError: If a list index is not a valid integer.
        """
        if isinstance(key, str):
            keys = key.split(".")
        else:
            keys = [key]

        target = data

        for k in keys[:-1]:
            if isinstance(target, dict):
                if create_missing:
                    target = target.setdefault(k, {})
                else:
                    if k not in target:
                        raise KeyError(f"Key '{k}' not found in dictionary.")
                    target = target[k]
            elif isinstance(target, list):
                index = self._validate_index(k)
                if create_missing:
                    # Extend the list to accommodate the index if out of range
                    if index >= len(target):
                        target.extend({} for _ in range(index - len(target) + 1))
                elif index >= len(target):
                    raise IndexError(f"Index {index} out of range for list.")
                target = target[index]
            else:
                raise TypeError(f"Cannot navigate through {type(target).__name__}.")

        return target, keys[-1]

    def items(self, data: Union[Dict, list]) -> Iterator:
        """Return an iterator over key-value or index-value pairs."""
        if isinstance(data, dict):
            return iter(data.items())
        elif isinstance(data, list):
            return iter(enumerate(data))
        else:
            raise TypeError("Items method only supports dicts and lists.")

    def replace_indirect(self, data: Dict, key: str) -> Union[str, None]:
        """
        Resolve indirect key references in a string.

        Args:
            data (Dict): Dictionary containing potential indirect references.
            key (str): The key to resolve.

        Returns:
            Union[str, None]: The resolved key, or None if the reference could not be resolved.

        Notes:
            - If the key is not a string, it is returned as is.
            - If the key contains '@' for indirect references, it tries to resolve the reference.
        """
        if not isinstance(key, str):
            return key

        if "@" in key:
            try:
                main_key, indirect_ref = key.split("@", 1)
            except ValueError:
                self.warn(f"Invalid indirect key format: '{key}'")
                return None

            ref = data.get(indirect_ref)
            if ref is not None:
                return f"{main_key}{ref}"
            else:
                self.warn(f"Indirect key reference '{indirect_ref}' not found in data.")
                return None

        return key

    def _validate_index(self, key: str) -> int:
        """
        Validate and convert a key to an integer index for list access.

        Args:
            key (str): The key to validate.

        Returns:
            int: The validated integer index.

        Raises:
            ValueError: If the key is not a valid integer.
        """
        try:
            return int(key)
        except ValueError:
            raise ValueError(f"Key '{key}' is not a valid integer index for list access.")

    def warn(self, message: str) -> None:
        """Print a warning message if verbosity is enabled."""
        if self.verbose > 2:
            print(f"Warning: {message}")



from unittest.mock import MagicMock

from PyQt6.QtWidgets import QApplication, QComboBox, QLineEdit, QTextEdit
import pytest

from YMLEditor.item_widget import ItemWidget


@pytest.fixture(scope="module")
def app():
    """
    Create a QApplication instance for testing.
    Required for PyQt widgets.
    """
    return QApplication([])


@pytest.fixture
def mock_config():
    """
    Mock config object with `get`, `set`, `save`, and `load` methods.
    """
    config = MagicMock()
    config.get.side_effect = lambda key: {"key1": "value1", "key2": "value2"}.get(key, None)
    return config


@pytest.fixture
def mock_callback():
    """
    Mock callback function.
    """
    return MagicMock()


def test_item_widget_combo(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="combo", initial_value="Option1",
        combo_rgx=["Option1", "Option2"], callback=mock_callback, key="key1", )
    assert isinstance(widget.widget, QComboBox)
    assert widget.widget.currentText() == "Option1"

def test_item_widget_line_edit(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="line_edit", initial_value="Initial Text", combo_rgx=r"^\w+$",
        # Regex for validation
        callback=mock_callback, key="key2", )
    assert isinstance(widget.widget, QLineEdit)
    assert widget.widget.text() == "Initial Text"
    assert widget._rgx == r"^\w+$"


def test_item_widget_text_edit(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="text_edit", initial_value="Multiline Text", combo_rgx=r".*",
        # Allow any text
        callback=mock_callback, text_edit_height=100, key="key3", )
    assert isinstance(widget.widget, QTextEdit)
    assert widget.widget.toPlainText() == "Multiline Text"
    assert widget.widget.height() == 100


def test_item_widget_read_only(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="read_only", initial_value="Read-Only Text", combo_rgx=None,
        callback=mock_callback, key="key4", )
    assert isinstance(widget.widget, QLineEdit)
    assert widget.widget.text() == "Read-Only Text"
    assert widget.widget.isReadOnly()


def test_item_widget_on_widget_changed_invalid(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="line_edit", initial_value="ValidText", combo_rgx=r"^\w+$",
        callback=mock_callback, key="key1", )
    line_edit = widget.widget
    line_edit.setText("Invalid Text!")  # Fails regex
    widget._on_widget_changed(line_edit)
    assert not widget._is_valid
    assert line_edit.styleSheet() == widget._error_style


def test_item_widget_set_text(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="line_edit", initial_value="Initial", combo_rgx=None,
        callback=mock_callback, key="key1", )
    widget.set_text(widget.widget, "Updated Value")
    assert widget.widget.text() == "Updated Value"

def test_item_widget_set_none(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="line_edit", initial_value="Initial", combo_rgx=None,
        callback=mock_callback, key="key1", )
    # 1) Set text
    widget.set_text(widget.widget, "Updated Value")
    assert widget.widget.text() == "Updated Value"
    # 2) Set None
    widget.set_text(widget.widget, None)
    assert widget.widget.text() == 'None'
    # 3) Set text
    widget.set_text(widget.widget, "Updated Value2")
    assert widget.widget.text() == "Updated Value2"


def test_item_widget_display(app, mock_config, mock_callback):
    widget = ItemWidget(
        config=mock_config, widget_type="line_edit", initial_value="", combo_rgx=None,
        callback=mock_callback, key="key1", )
    widget.display()
    assert widget.widget.text() == "value1"  # From mock_config.get()
